import gw_eccentricity
from gw_eccentricity import load_data
from gw_eccentricity import measure_eccentricity
import numpy as np


def test_time_convention():
    """ Tests that the measure_eccentricity interface is working for any
    time conventions.

    We generate a dataDict using EccentricTD waveform and then create
    two different copy of the dataDict with
    - the fist dataDict time having t=0 at the begining and
    - the second dataDict time having t=0 at the end

    The test expects that the measured eccentricity array and mean anomaly
    array from these two dataDicts should be the same.
    """
    # Load test waveform
    lal_kwargs = {"approximant": "EccentricTD",
                  "q": 3.0,
                  "chi1": [0.0, 0.0, 0.0],
                  "chi2": [0.0, 0.0, 0.0],
                  "Momega0": 0.01,
                  "ecc": 0.1,
                  "mean_ano": 0,
                  "include_zero_ecc": True}
    dataDict = load_data.load_waveform(**lal_kwargs)

    # create dataDict with t=0 as the begining
    dataDict1 = dataDict.copy()
    dataDict1["t"] = dataDict1["t"] - dataDict1["t"][0]
    dataDict1["t_zeroecc"] = (dataDict1["t_zeroecc"]
                              - dataDict1["t_zeroecc"][0])

    # create dataDict with t=0 as the end
    dataDict2 = dataDict.copy()
    dataDict2["t"] = dataDict2["t"] - dataDict2["t"][-1]
    dataDict2["t_zeroecc"] = (dataDict2["t_zeroecc"]
                              - dataDict2["t_zeroecc"][-1])

    # time shift between dataDict1 and dataDict2
    time_shift = dataDict1["t"][0] - dataDict2["t"][0]

    # List of all available methods
    available_methods = gw_eccentricity.get_available_methods()
    for method in available_methods:
        eccs = []
        meananos = []
        tref_outs = []
        for data in [dataDict1, dataDict2]:
            gwecc_dict = measure_eccentricity(
                tref_in=data["t"],
                method=method,
                dataDict=data)
            tref_out = gwecc_dict["tref_out"]
            ecc_ref = gwecc_dict["eccentricity"]
            meanano_ref = gwecc_dict["mean_anomaly"]
            tref_outs.append(tref_out)
            eccs.append(ecc_ref)
            meananos.append(meanano_ref)
        np.testing.assert_allclose(tref_outs[0], tref_outs[1] + time_shift)
        np.testing.assert_allclose(eccs[0], eccs[1])
        np.testing.assert_allclose(np.unwrap(meananos[0]),
                                   np.unwrap(meananos[1]))



"""
Find peaks and troughs using amplitude fits.

Part of Eccentricity Definition project.
"""
from .eccDefinitionUsingFrequencyFits import eccDefinitionUsingFrequencyFits
from .plot_settings import labelsDict
from .utils import check_kwargs_and_set_defaults


class eccDefinitionUsingAmplitudeFits(eccDefinitionUsingFrequencyFits):
    """Measure eccentricity by finding extrema location using amp fits."""

    def __init__(self, *args, **kwargs):
        """Init for eccDefinitionUsingWithAmplitudeFits class.

        parameters:
        ----------
        dataDict: Dictionary containing the waveform data.
        """
        super().__init__(*args, **kwargs)
        self.data_str = "amp_gw"
        self.label_for_data_for_finding_extrema = self.get_label_for_plots("amp")
        self.label_for_fit_to_data_for_finding_extrema \
            = labelsDict[f"{self.data_str}_fit"]
        self.method = "AmplitudeFits"
        self.kwargs_for_fits_methods = check_kwargs_and_set_defaults(
            self.extra_kwargs['kwargs_for_fits_methods'],
            self.get_default_kwargs_for_fits_methods(),
            "kwargs_for_fits_methods",
            "eccDefinitionUsingAmplitudeFits.get_default_kwargs_for_fits_methods()")
        self.set_fit_variables()
        # Make a copy of amp_gw and use it to set data_for_finding_extrema.
        # This would ensure that any modification of data_for_finding_extrema
        # does not modify amp_gw.
        self.data_for_finding_extrema = self.amp_gw.copy()
        # FIXME: Find a better solution
        # It turns out that since in MKS units amplitude is very small
        # The envelope fitting does not work properly. Maybe there is a better
        # way to do this but scaling the amp_gw data by its value at the global
        # peak (the merger time) solves this issue.
        self.data_for_finding_extrema /= self.amp_gw_merger

    def get_default_kwargs_for_fits_methods(self):
        """Get default fits kwargs.

        See eccDefinitionUsingFrequencyFits.get_default_kwargs_for_fits_methods
        for documentation.
        """
        return {
            # The PN exponent as approximation
            # Jolien and Creighton Chapter 3, Equation 3.190a, 3.190b
            "nPN": -1./4,
            "fit_bounds_max_amp_factor": 10,
            "fit_bounds_max_nPN_factor": 10,
            "prominence_factor": 0.025,
            "distance_factor": 0.5,
            "num_orbits": 3,
            "num_orbits_for_global_fit": 10,
            "return_diagnostic_data": False
        }



"""Compare different eccDefinition methods."""

import numpy as np


def compute_errors_between_methods(gwecc_obj1,
                                   gwecc_obj2,
                                   tmin=None,
                                   tmax=None):
    """Compute errors in eccentricity and mean anomaly from two methods.

    This function computes the errors (difference) in the measured value of
    eccentricity and mean anomaly using two different methods. Since both
    methods might be not able to measure eccentricity and mean anomaly at
    the same range of times, this function will return the errors in the region
    of common times.

    Parameters:
    -----------
    gwecc_obj1:
        gwecc_object using method 1 as returned by
        gw_eccentricity.measure_eccentricity function with
        "return_gwecc_object" set to True.
    gwecc_obj2:
        gwecc_object using method 2 as returned by
        gw_eccentricity.measure_eccentricity function with
        "return_gwecc_object" set to True.
    tmin:
        If not None, errors are computed only for times later than tmin.
        If None, errors are computed for times later than
        max(gwecc_obj1.tmin, gwecc_obj2.tmin).
        Default is None.
    tmax:
        If not None, errors are computed only for times earlier than tmax.
        If None, errors are computed for times earlier than
        min(gwecc_obj1.tmax, gwecc_obj2.tmax).
        Default is None.

    Returns:
    -------
    t:
        Times where errors are computed. The minimum/maximum t is choosen
        such that the times in t are common to both methods. That is,
        tmin = max(gwecc_obj1.tmin, gwecc_obj2.tmin) and
        tmax = min(gwecc_obj1.tmax, gwecc_obj2.tmax).
        If in addition, tmin/tmax is not None, then t is further truncated
        such that t lies within tmin/tmax.
    ecc_errors:
        Absolute errors in eccentricity measured by method 1 (gwecc_obj1) and
        method 2 (gwecc_obj2).
    mean_ano_errors:
        Absolute errors in mean anomaly measured by method 1 (gwecc_obj1) and
        method 2 (gwecc_obj2).
    ecc_ref_1_common:
        Eccentricities at t measured using method 1.
    ecc_ref_2_common:
        Eccentricities at t measured using method 2.
    mean_ano_ref_1_common:
        Mean anomalies at t measured using method 1.
    mean_ano_ref_2_common:
        Mean anomalies at t measured using method 2.
    """
    # Check that the gwecc objects were created using same tref_in
    np.testing.assert_allclose(gwecc_obj1.tref_in, gwecc_obj2.tref_in)
    # Get the bounds for times within which both methods work
    tMinCommon = max(gwecc_obj1.tmin, gwecc_obj2.tmin)
    tMaxCommon = min(gwecc_obj1.tmax, gwecc_obj2.tmax)

    # Reset tMinCommon/tMaxCommon if tmin/tmax is provided
    if tmin is not None:
        if tmin > tMaxCommon:
            raise Exception(f"No common time found later than {tmin}")
        tMinCommon = max(tMinCommon, tmin)
    if tmax is not None:
        if tmax < tMinCommon:
            raise Exception(f"No common time found earlier than {tmax}")
        tMaxCommon = min(tMaxCommon, tmax)

    # Get indices for the common tref_out
    common_idx1 = np.logical_and(gwecc_obj1.tref_out >= tMinCommon,
                                 gwecc_obj1.tref_out < tMaxCommon)
    common_idx2 = np.logical_and(gwecc_obj2.tref_out >= tMinCommon,
                                 gwecc_obj2.tref_out < tMaxCommon)
    tref_out_1_common = gwecc_obj1.tref_out[common_idx1]
    tref_out_2_common = gwecc_obj2.tref_out[common_idx2]
    # Check that common tref out is the same.
    np.testing.assert_allclose(tref_out_1_common, tref_out_2_common)
    # Compute errors in eccentricity at common tref_out
    ecc_ref_1_common = gwecc_obj1.eccentricity[common_idx1]
    ecc_ref_2_common = gwecc_obj2.eccentricity[common_idx2]
    ecc_errors = np.abs(ecc_ref_1_common - ecc_ref_2_common)
    # Compute errors in mean anomaly at common tref_out
    mean_ano_ref_1_common = gwecc_obj1.mean_anomaly[common_idx1]
    mean_ano_ref_2_common = gwecc_obj2.mean_anomaly[common_idx2]
    # We need to unwrap the mean anomaly since zero and
    # 2pi should be treated as the same and hence should give zero errors.
    mean_ano_errors = np.abs(
        np.unwrap(mean_ano_ref_1_common)
        - np.unwrap(mean_ano_ref_2_common))
    return tref_out_1_common, ecc_errors, mean_ano_errors, ecc_ref_1_common,\
        ecc_ref_2_common, mean_ano_ref_1_common, mean_ano_ref_2_common



"""Useful functions for the project."""
import numpy as np
import argparse
from polyrat import StabilizedSKRationalApproximation
from scipy.interpolate import InterpolatedUnivariateSpline
from scipy.interpolate import PchipInterpolator
from scipy.optimize import least_squares
import warnings


def amplitude_using_all_modes(mode_dict, data_type="hlm"):
    """Get the amplitude using all the available modes.

    Parameters
    ----------
    mode_dict: dict
        Dictionary containing waveform modes.
    data_type: str
        Indicate whether the dictionary contains hlm or amplm.
        Default is "hlm".

    Returns
    -------
        Square root of the qudrature sum of the amplitudes of all the
        available modes in mode_dict.
    """
    amp = 0
    for mode in mode_dict.keys():
        if data_type == "hlm":
            amp += np.abs(mode_dict[mode])**2
        elif data_type == "amplm":
            amp += mode_dict[mode]**2
        else:
            raise KeyError("data_type must be either `hlm` or `amplm`.")
    return np.sqrt(amp)


def peak_time_via_quadratic_fit(t, func):
    """
    Find the peak time of a function quadratically.

    Fits the function to a quadratic over the 5 points closest to the argmax
    func.

    Parameters
    ----------
    t:
        An array of times.
    func:
        Array of function values.

    Returns
    -------
    tpeak:
        Time at peak of the function func.
    fpeak:
        Value of function func at tpeak.
    """
    # Find the time closest to the peak, making sure we have room on either
    # side
    index = np.argmax(func)
    index = max(2, min(len(t) - 3, index))

    # Do a quadratic fit to 5 points,
    # subtracting t[index] to make the matrix inversion nice
    testTimes = t[index-2:index+3] - t[index]
    testFuncs = func[index-2:index+3]
    xVecs = np.array([np.ones(5), testTimes, testTimes**2.])
    invMat = np.linalg.inv(np.array([[v1.dot(v2) for v1 in xVecs]
                                     for v2 in xVecs]))

    yVec = np.array([testFuncs.dot(v1) for v1 in xVecs])
    coefs = np.array([yVec.dot(v1) for v1 in invMat])
    return t[index] - coefs[1]/(2.*coefs[2]), (coefs[0]
                                               - coefs[1]**2./4/coefs[2])


def check_kwargs_and_set_defaults(user_kwargs=None,
                                  default_kwargs=None,
                                  name="user given kwargs",
                                  location=None):
    """Sanity check user given dicionary of kwargs and set default values.

    Parameters
    ----------
    user_kwargs:
        Dictionary of kwargs by user.
    default_kwargs:
        Dictionary of default kwargs.
    name:
        string to represent the dictionary
    location:
        string pointing to where the defaults are defined

    Returns
    -------
    updated user_kwargs
    """
    # make user_kwargs iterable
    if user_kwargs is None:
        user_kwargs = {}

    for kw in user_kwargs.keys():
        if kw not in default_kwargs:
            raise ValueError(f"Invalid key {kw} in {name}."
                             " Should be one of "
                             f"{list(default_kwargs.keys())}\n"
                             f"To add a new keyword, please modify {location}")

    for kw in default_kwargs.keys():
        if kw not in user_kwargs:
            user_kwargs[kw] = default_kwargs[kw]

    return user_kwargs


def raise_exception_if_none(kwargs, keys_to_check, name, location):
    """Raise exception if any key from `keys_to_check` has value None."""
    for kw in keys_to_check:
        if kwargs[kw] is None:
            raise Exception(f"kw {kw} for {name} can not be None."
                            f" Check documentation of {location} for more"
                            " details.")


class SmartFormatter(argparse.ArgumentDefaultsHelpFormatter):
    """Smart Formatter for argparse helper strings.

    Stolen from https://stackoverflow.com/questions/3853722/how-to-insert-newlines-on-argparse-help-text.
    """

    def _split_lines(self, text, width):
        if text.startswith('R|'):
            return text[2:].splitlines()
        # this is the RawTextHelpFormatter._split_lines
        return argparse.HelpFormatter._split_lines(self, text, width)


def time_deriv_4thOrder(y, dt):
    """Fourth order accurate time derivative.

    Assuming constant time step.
    Tested for convergence up to 1e-12 level.

    Parameters
    ----------
    y:
        1d array to take time derivative of.
    dt:
        Time step.

    Returns
    -------
    dydt:
        Fourth order time derivative of y.
    """
    # Use a 5 point stencil
    res = 0*y
    # First do the interior
    res[2:-2] = (y[:-4] - 8*y[1:-3] + 8*y[3:-1] - y[4:])/12.

    # Next do the edges
    res[1] = y[:5].dot(np.array([-3, -10, 18, -6, 1]) / 12.)
    res[0] = y[:5].dot(np.array([-25, 48, -36, 16, -3]) / 12.)
    res[-2] = y[-5:].dot(np.array([-1, 6, -18, 10, 3]) / 12.)
    res[-1] = y[-5:].dot(np.array([3, -16, 36, -48, 25]) / 12.)
    dydt = res / dt
    return dydt


def get_default_spline_kwargs():
    """Defaults for spline settings."""
    default_spline_kwargs = {
        "w": None,
        "bbox": [None, None],
        "k": 3,
        "ext": 2,
        "check_finite": False}
    return default_spline_kwargs


def interpolate(newX,
                oldX,
                oldY,
                allowExtrapolation=False,
                interpolator="spline",
                spline_kwargs=None,
                check_kwargs=True):
    """Interpolate.

    Parameters
    ----------
    newX:
        Points where interpolant is to be evaluated.
    oldX:
        1d array of monotonically increasing real values.
    oldY:
        1d array of monotonically increasing real values.
    allowExtrapolation:
        Bool. If True returns extrapolated values. Default is False.
        If False, an exception is raised if trying to extrapolate.
    interpolator:
        String to choose an interpolator to interpolate oldX, oldY.
        Could be one of the following:
        "spline": Uses scipy.interpolate.InterpolatedUnivariateSpline.
        "monotonic_spline":  Uses scipy.interpolate.PchipInterpolator.
        Default is "spline".
    spline_kwargs:
        See under get_interpolant.
    check_kwargs:
        Check spline_kwargs if check_kwargs is True. Default is True.

    Returns
    -------
    newY:
        Intepolated values at newX.
    """
    if len(oldY) != len(oldX):
        raise Exception("Lengths dont match.")

    if not allowExtrapolation:
        if np.min(newX) < np.min(oldX) - 1e-10 \
           or np.max(newX) > np.max(oldX) + 1e-10:
            print(f"Min of newX = {np.min(newX)}, "
                  f"Min of oldX = {np.min(oldX)}")
            print(f"Max of newX = {np.max(newX)}, "
                  f"Max of oldX = {np.max(oldX)}")
            print("newX has vlaues below oldX? "
                  f"{np.min(newX) < np.min(oldX)}")
            print("newX has values above oldX? "
                  f"{np.max(newX) > np.max(oldX)}")
            raise Exception("Trying to extrapolate, "
                            "but allowExtrapolation=False")
    newY = get_interpolant(oldX, oldY, allowExtrapolation, interpolator,
                           spline_kwargs, check_kwargs)(newX)
    return newY


def get_interpolant(oldX,
                    oldY,
                    allowExtrapolation=False,
                    interpolator="spline",
                    spline_kwargs=None,
                    check_kwargs=True):
    """Create Interpolant.

    Parameters
    ----------
    oldX:
        1d array of monotonically increasing real values.
    oldY:
        1d array of monotonically increasing real values.
    allowExtrapolation:
        Bool. If True returns extrapolated values. Default is False.
        If False, an exception is raised if trying to extrapolate.
    interpolator:
        String to choose an interpolator to interpolate oldX, oldY.
        Could be one of the following:
        "spline": Uses scipy.interpolate.InterpolatedUnivariateSpline.
        "monotonic_spline":  Uses scipy.interpolate.PchipInterpolator.
        Default is "spline".
    spline_kwargs:
        Dictionary of kwargs to be provided to the interpolator if
        "interpolator"="spline". The allowed kwargs are the same as that of the
        spline function scipy.interpolate.InterpolatedUnivariateSpline and the
        defaults are set using gw_eccentricity.utils.get_default_spline_kwargs.
        Since we use allowExtraplotion arg separately, value of "ext" in
        extra_kwargs will be overridden by allowExtrapolation.
    check_kwargs:
        Check spline_kwargs if check_kwargs is True. Default is True.

    Returns
    -------
    Intepolatnt.
    """
    if not np.all(np.diff(oldX) > 0):
        raise Exception("oldX must have increasing values")

    if interpolator == "spline":
        if check_kwargs:
            spline_kwargs = check_kwargs_and_set_defaults(
                spline_kwargs, get_default_spline_kwargs(), "spline kwargs",
                "utils.get_default_spline_kwargs")
        # check that num of data points > order of spline
        if len(oldX) >= 2:
            if len(oldX) <= spline_kwargs["k"]:
                warnings.warn(f"No of data points is {len(oldX)} but "
                              f"spline order k = {spline_kwargs['k']}. "
                              f"Decreasing k to {len(oldX) - 1}.")
                # make a copy so that the original spline_kwargs remains
                # unmodified.
                kwargs = spline_kwargs.copy()
                kwargs["k"] = len(oldX) - 1
            else:
                kwargs = spline_kwargs
        else:
            raise Exception("Number of data points is {len(oldX)}."
                            " Cannot build an interpolant.")
        # If allowExtrapolation is True but ext=2 then raise a warning
        # and override ext to 0
        if allowExtrapolation and kwargs["ext"] == 2:
            # ext = 0, returns extraploted values
            kwargs["ext"] = 0
        # If allowExtraplotion is False but ext != 2 then raise a warning
        # and override ext to 2
        if not allowExtrapolation and kwargs["ext"] != 2:
            # ext = 2, raises exception if extrapolation is attempted.
            kwargs["ext"] = 2
        interpolant = InterpolatedUnivariateSpline(oldX, oldY, **kwargs)
    elif interpolator == "monotonic_spline":
        if spline_kwargs is not None:
            warnings.warn(f"Interpolator is {interpolator} but spline_kwargs "
                          "are passed. spline_kwargs will be ignored.")
        interpolant = PchipInterpolator(oldX, oldY,
                                        extrapolate=allowExtrapolation)
    else:
        raise ValueError(f"Unknown interpolator {interpolator}. Must be one"
                         " of ['spline', 'monotonic_spline']")
    return interpolant


def get_default_rational_fit_kwargs():
    """Get default kwargs for rational fit."""
    default_rational_fit_kwargs = {
        "num_degree": None,
        "denom_degree": None,
        "norm": 2,
        "maxiter": 20,
        "verbose": False,
        "xtol": 1e-07,
    }
    return default_rational_fit_kwargs


def get_rational_fit(x, y, rational_fit_kwargs=None, check_kwargs=True):
    """Get rational fit for the data set (x, y).

    We use `polyrat.StabilizedSKRationalApproximation` to obtain
    rational fit to the data.
    """
    if check_kwargs:
        rational_fit_kwargs = check_kwargs_and_set_defaults(
            rational_fit_kwargs,
            get_default_rational_fit_kwargs(),
            "rational_fit_kwargs",
            "utils.get_default_rational_fit_kwargs"
        )
    # We use a rational approximation based on Stabilized Sanathanan-Koerner Iteration
    # described in arXiv:2009.10803 and implemented in `polyrat.StabilizedSKRationalApproximation`.
    rat = StabilizedSKRationalApproximation(**rational_fit_kwargs)
    # The input x-axis data must be 2-dimensional
    rat.fit(x.reshape(-1, 1), y)
    # Using a lambda function to change the input 1-d data to
    # float64 and reshape to 2-d as required by the fit.
    # Ensure also that the input and output types are the same.
    return lambda t: (rat(np.atleast_2d(t).astype('float64').reshape(-1, 1)).item() if np.isscalar(t)\
                      else rat(np.atleast_2d(t).astype('float64').reshape(-1, 1)))


def debug_message(message, debug_level, important=True,
                  point_to_verbose_output=False):
    """Show message based on debug_level.

    parameters
    ----------
    message: str
        Message to display.

    debug_level: int
        Indicator for level of debug message. Based on it, one of the
        following actions if performed:
        -1: No action is performed and hence no message is displayed.
        0: Warning is issued with the input message only if important=True
        1: Warning is issued with the input message.
        2: Exception is raised with the input message.

    important: bool
        Only if True, the message gets printed when debug_level=0. For
        other debug_levels, this does nothing.
        Default is True.

    point_to_verbose_output: bool
        When True, if debug_level is 0 and important is True, points to
        debug_level = 1 for more verbose output.
        Default is False
    """
    debug_levels = [-1, 0, 1, 2]
    if debug_level not in debug_levels:
        raise ValueError(
            f"Unknown debug_level {debug_level}. Should one "
            f"of {debug_levels}. See "
            "`gw_eccentricity.utils.debug_message` for action"
            "performed with each debug level.")
    if debug_level == -1:
        # Do nothing
        return
    if (debug_level == 0 and important) or debug_level == 1:
        if debug_level == 0 and point_to_verbose_output:
            message += "\nFor more verbose output use `debug_level=1`."
        # Issue warning. Use stacklevel=2 to point to actual line number
        # causing this warning instead of pointing to here.
        warnings.warn(message, stacklevel=2)
    if debug_level == 2:
        # raise Exception
        raise Exception(message)

class QusicircularFitForMappingFrequencyToTime:
    """Fit eccentric frequency to qausicircular frequency.
    
    Fitting nonmonotonic eccentric frequency to quasicircular monotonic
    frequency provides an inexpensive way to obtain an approximate map between
    a given orbit averaged frequency and the corresponding time in the
    evolution of the binary. This is useful when we want the frequency to time
    map without actually constructing an orbit averaged frequency which
    requires one to find the extrema in the eccentric frequency first.
    """

    def __init__(self, times, frequencies):
        """Init."""
        self.times = times
        self.frequencies = frequencies
        self.t0 = 0.5 * (times[0] + times[-1])

    def residual(self, p, t, w):
        f0, f1, T = p
        return w - self.model(t, f0, f1, T)

    def model(self, t, f0, f1, T):
        """model for quasicircular frequency."""
        # f0, f1 are function values and first time-derivatives
        # at t0.  Re-expfress as T, n, A, then evalate A*(T-t)^n
        n = -(T - self.t0) * (f1/f0)
        A = f0 * (T - self.t0)**(-n)
        return A * (T - t)**n

    def least_square_fit(self):
        # typial scale of data
        f0 = 0.5 * (max(self.frequencies) + min(self.frequencies)) #0.5 * (self.frequencies[0] + self.frequencies[-1])
        p0 = [f0,  # function value
              (3.0/8.0) * f0 / (-self.t0),  # func = f0/t0^n*(t)^n -> dfunc/dt (t0) = n*f0/t0
              0  # singularity in fit is near t=0, since waveform aligned at max(amp_gw)
              ]
        bounds = [[0., 0., 0.8*self.times[-1]],
                  [10 * f0,
                   10 * f0 / (-self.t0),
                   -self.t0]]
        res = least_squares(self.residual, p0, bounds=bounds, args=(self.times, self.frequencies),
                            method='trf', max_nfev=100)
        return res



"""
Find peaks and troughs using frequency fits.

Part of Eccentricity Definition project.
"""
from .eccDefinition import eccDefinition
from .plot_settings import labelsDict
import numpy as np
import scipy
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from .utils import check_kwargs_and_set_defaults


class envelope_fitting_function:
    """Fit envelope.

    Re-parameterize A*(T-t)^n in terms of function value and 
    first derivative at the time t0, and T.
    """

    def __init__(self, t0, verbose=False):
        """Init."""
        self.t0 = t0
        self.verbose = verbose

    def format(self, f0, f1, T):
        """Return a string representation for use in legends and output."""
        n = -(T-self.t0)*f1/f0
        A = f0*(T-self.t0)**(-n)
        return f"{A:.3g}({T:+.2f}-t)^{n:.3f}"

    def __call__(self, t, f0, f1, T):
        """Call."""
        # f0, f1 are function values and first time-derivatives
        # at t0.  Re-expfress as T, n, A, then evalate A*(T-t)^n
        n = -(T-self.t0)*f1/f0
        A = f0*(T-self.t0)**(-n)
        if self.verbose:
            print(f"f0={f0}, f1={f1}, T={T}; n={n}, A={A}, max(t)={t.max()}")
        if t.max() > T:
            print(end="", flush=True)
            raise Exception(
                "envelope_fitting_function reached parameters where merger "
                "time T is within time-series to be fitted\n"
                f"f0={f0}, f1={f1}, T={T}; n={n}, A={A}, max(t)={max(t)}")
        return A*(T-t)**n


class eccDefinitionUsingFrequencyFits(eccDefinition):
    """Measure eccentricity by finding extrema location using freq fits."""

    def __init__(self, *args, **kwargs):
        """Init for eccDefinitionUsingWithFrequencyFits class.

        Parameters
        ----------
        dataDict : dict
            Dictionary containing the waveform data.
        """
        super().__init__(*args, **kwargs)
        self.data_str = "omega_gw"
        self.label_for_data_for_finding_extrema = self.get_label_for_plots("omega")
        self.label_for_fit_to_data_for_finding_extrema \
            = labelsDict[f"{self.data_str}_fit"]
        # Make a copy of omega_gw and use it to set data_for_finding_extrema.
        # This would ensure that any modification of data_for_finding_extrema
        # does not modify omega_gw.
        self.data_for_finding_extrema = self.omega_gw.copy()
        self.method = "FrequencyFits"
        # Get dictionary of kwargs to be used for Fits methods.
        self.kwargs_for_fits_methods = check_kwargs_and_set_defaults(
            self.extra_kwargs['kwargs_for_fits_methods'],
            self.get_default_kwargs_for_fits_methods(),
            "kwargs_for_fits_methods",
            "eccDefinitionUsingFrequencyFits.get_default_kwargs_for_fits_methods()")
        # Set variables needed for envelope fits and find_peaks
        self.set_fit_variables()
        # show more verbose output if debug_level is >= 1
        self.verbose = self.debug_level >= 1
        # If return_diagnostic_data is true then return a dictionary of data for diagnostics.
        if self.return_diagnostic_data:
            self.diagnostic_data_dict = {
                # Initialize with empty lists to fill in during iteration over the extrema
                "params": {"pericenters": [], "apocenters": []},
                "t_extrema": {"pericenters": [], "apocenters": []},
                "data_extrema": {"pericenters": [], "apocenters": []},
                "t_ref": {"pericenters": [], "apocenters": []},
                "f_fit": {"pericenters": [], "apocenters": []},
                "data": self.data_for_finding_extrema,
                "t": self.t
            }

    def get_default_kwargs_for_fits_methods(self):
        """Get default kwargs to be used for Fits methods.

        The kwargs are

        - "nPN": The PN exponent to use in the fit function. It is
          inspired by the functional form of frequency/amplitude in
          the leading Post-Newtonian order ~(t - t_merger)^nPN
        - "fit_bounds_max_amp_factor": To set the upper bound on the
          Amplitude A of the fitting function of the form A(t-T)^n.
          The upper bound of A is set as f0*fit_bounds_max_amp_factor,
          where f0 is the mean of the first and the last values of data
          to be fitted.  f0 = 0.5*(data[0]+data[-1])
        - "fit_bounds_max_nPN_factor": To set the upper bound on the
          exponent n of the fitting function. The upper bound on n is
          set as f0*fit_bounds_max_nPN_factor/(-fit_center_time),
          where fit_center_time is the time at the midpoint of the
          data.
          fit_center_time = 0.5*(t[0]+t[-1]).
          The merger is assumed to be at t=0 here.
        - "prominence_factor": To set the prominence for find_peaks
          function. The prominence is set as
          prominence = prominence_factor * residual_data_amp,
          where,
          residual_data_amp = max(residual_data) - min(residual_data)
        - "distance_factor": To set the distance for find_peaks
          function.
          distance = distance_factor * average_orbital_period
        - "num_orbits": Number of extrema to look for during
          fitting. It looks for num_orbits on the left and num_orbits+1
          on the right.
        - "num_orbits_for_global_fit": Number of orbits to use for
          global fit in the Fits methods.
        - "return_diagnostic_data": If True, Retuns a dictionary
          of data useful for diagnostics. Default is False.
        """
        return {
            "nPN": -3./8,
            "fit_bounds_max_amp_factor": 10,
            "fit_bounds_max_nPN_factor": 10,
            "prominence_factor": 0.03,  # prominence = 3% of residual_amp_max
            "distance_factor": 0.75,  # 75% of the average orbital period,
            "num_orbits": 3,
            "num_orbits_for_global_fit": 10,
            "return_diagnostic_data": False
        }

    def set_fit_variables(self):
        """Set variables to be used for Fits Methods.

        See under get_default_kwargs_for_fits_methods for documentation
        on these variables.
        """
        self.fit_bounds_max_amp_factor = self.kwargs_for_fits_methods[
            "fit_bounds_max_amp_factor"]
        self.fit_bounds_max_nPN_factor = self.kwargs_for_fits_methods[
            "fit_bounds_max_nPN_factor"]
        self.nPN = self.kwargs_for_fits_methods["nPN"]
        self.prominence_factor = self.kwargs_for_fits_methods[
            "prominence_factor"]
        self.distance_factor = self.kwargs_for_fits_methods["distance_factor"]
        self.num_orbits = self.kwargs_for_fits_methods["num_orbits"]
        self.num_orbits_for_global_fit = self.kwargs_for_fits_methods[
            "num_orbits_for_global_fit"]
        self.return_diagnostic_data = self.kwargs_for_fits_methods["return_diagnostic_data"]

    def find_extrema(self, extrema_type="pericenters"):
        """Find the extrema in the data.

        Parameters
        ----------
        extrema_type:
            Either "pericenters" or "apocenters".

        Returns
        -------
        array of positions of extrema.
        """
        # STEP 0 - setup

        if extrema_type == "pericenters":
            sign = +1
        elif extrema_type == "apocenters":
            sign = -1
        else:
            raise Exception(f"extrema_type='{extrema_type}' unknown.")
        
        # The fit function assume the merger to be at t=0. So we align the time
        # axis such that merger occurs at t=0.
        # After the peaks are found, the time is reshifted to to its original
        # values at the end of this function so the merger is again at t_merger.
        self.t -= self.t_merger

        # DESIRED NUMBER OF EXTREMA left/right DURING FITTING
        # Code will look for N extrema to the left of idx_ref, and N+1 extrema
        # to the right
        N = self.num_orbits

        # if True, perform an additional fitting step to find the position of
        # extrema to sub-gridspacing accuracy.
        #
        # The results of this more accurate extrema-determination are collected
        # in self.periastron_info = [t_extrema_refined,
        # data_extrema_refined, phase_gw_extrema_refined] and/or
        # self.apastron_info = [t_extrema_refined, data_extrema_refined,
        # phase_gw_extrema_refined]
        # if False, then
        # self.periastron_info/apastron_info contains the data at the
        # grid-points
        refine_extrema = self.extra_kwargs["refine_extrema"]

        # diagnostic output?
        # setting diag_file to a valid pdf-filename will trigger diagnostic
        # plots
        # Print more verbose output if debug_level >= 1
        # Create debug plots if debug_plots is True
        diag_file = (f"gwecc_{self.method}_diagnostics_{extrema_type}.pdf") if self.debug_plots else ""
        pp = PdfPages(diag_file) if diag_file != "" else False
        # STEP 1:
        # global fit as initialization of envelope-subtraced extrema

        # use this many orbits from the start of the waveform for the initial
        # global fit.
        # Keeping this initial fit-interval away from merger helps
        # to obtain a good fit that also allows to discern small eccentricities
        N_orbits_for_global_fit = self.num_orbits_for_global_fit
        idx_end = np.argmax(self.phase_gw > self.phase_gw[0]
                            + N_orbits_for_global_fit*4*np.pi)

        if idx_end == 0:  # don't have that much data, so use all
            idx_end = -1

        if self.verbose:
            print(f"t[0]={self.t[0]}, t[-1]="
                  f"{self.t[-1]}, "
                  f"global fit to t<={self.t[idx_end]}")

        # create fitting function object, set initial guess and bounds
        fit_center_time = 0.5*(self.t[0] + self.t[-1])
        f_fit = envelope_fitting_function(t0=fit_center_time,
                                          verbose=False)
        # typial scale of data
        f0 = 0.5 * (self.data_for_finding_extrema[0]+self.data_for_finding_extrema[idx_end])
        p0 = [f0,  # function value
              -self.nPN*f0/(-fit_center_time),  # func = f0/t0^n*(t)^n -> dfunc/dt (t0) = n*f0/t0
              0  # singularity in fit is near t=0, since waveform aligned at max(amp_gw)
              ]
        bounds0 = [[0., 0., 0.8*self.t[-1]],
                   [self.fit_bounds_max_amp_factor*f0,
                    self.fit_bounds_max_nPN_factor*f0/(-fit_center_time),
                    -fit_center_time]]
        if self.verbose:
            print(f"global fit: guess p0={p0},  t_center={fit_center_time}")
            print(f"            bounds={bounds0}")

        if pp:
            # first diagnostic plots.  Will be available even if
            # scipy.optimize fails
            fig, axs = plt.subplots(1, 3, figsize=(12, 4))
            axs[0].set_title(self.label_for_data_for_finding_extrema)
            axs[1].set_title(
                f"residual:  {self.label_for_data_for_finding_extrema}"
                f"-{self.label_for_fit_to_data_for_finding_extrema} "
                "(fitted region only)")
            axs[2].set_title(
                f"residual:  {self.label_for_data_for_finding_extrema}-"
                f"{self.label_for_fit_to_data_for_finding_extrema}")
            axs[0].plot(self.t, self.data_for_finding_extrema,
                        label=self.label_for_data_for_finding_extrema)
            axs[0].plot(
                self.t, f_fit(self.t, *p0),
                linewidth=0.5, color='grey',
                label=(f"{self.label_for_fit_to_data_for_finding_extrema}"
                       "(*p_guess)"))

        p_global, pconv = scipy.optimize.curve_fit(
            f_fit, self.t[:idx_end],
            self.data_for_finding_extrema[:idx_end], p0=p0,
            bounds=bounds0)
        if self.verbose:
            print(f"            result p_global={p_global}")

        if pp:
            line, = axs[0].plot(self.t[:idx_end],
                                f_fit(self.t[:idx_end], *p_global),
                                label='fit [first 10 orbits]')
            axs[0].plot(self.t, f_fit(self.t, *p_global),
                        color=line.get_color(), linewidth=0.5)

            line, = axs[1].plot(
                self.t[:idx_end],
                self.data_for_finding_extrema[:idx_end]
                - f_fit(self.t[:idx_end], *p_global),
                label='residual (fitted region only)')
            line, = axs[2].plot(
                self.t[:idx_end],
                self.data_for_finding_extrema[:idx_end]
                - f_fit(self.t[:idx_end], *p_global),
                label='residual (fitted region)')
            axs[2].plot(self.t,
                        self.data_for_finding_extrema-f_fit(self.t, *p_global),
                        linewidth=0.5, color=line.get_color(),
                        label='residual (all data)')
            axs[0].legend()
            axs[2].legend()
            fig.tight_layout()
            # fig.savefig(pp, format='pdf')
            self.save_debug_fig(fig, pp, diag_file)
            plt.close(fig)

        # STEP 2 From start of data, move through data and do local fits across
        # (2N+1) extrema.  For each fit, take the middle one as part of the
        # output

        # collects indices of extrema
        extrema = []

        # collects floating point values associated with extrema
        # if refine_extrema==true, these will in general be betweem
        # grid-points
        t_extrema_refined = []
        data_extrema_refined = []
        phase_gw_extrema_refined = []

        # estimates for initial start-up values. Because idx_ref
        # is only allowed to increase, be rather conservative with
        # its initial guess
        K = 1.1   # periastron-advance rate
        idx_ref = np.argmax(self.phase_gw
                            > self.phase_gw[0] + K*(N-1)*4*np.pi)
        if idx_ref == 0:
            raise Exception("data set too short.")
        p = p_global
        count = 0
        extra_extrema = []
        extra_extrema_t = []
        extra_extrema_data = []
        extra_extrema_phase_gw = []
        while True:
            count = count+1
            if self.verbose:
                print(f"=== count={count} "+"="*60)
            idx_extrema, p, K, idx_ref, extrema_refined \
                = self.FindExtremaNearIdxRef(
                    idx_ref,
                    sign, N, N+1, K,
                    f_fit, p, bounds0,
                    1e-8,
                    increase_idx_ref_if_needed=True,
                    refine_extrema=refine_extrema,
                    verbose=self.verbose,
                    pp=pp,
                    plot_info=f"count={count}")
            if self.verbose:
                print(f"IDX_EXTREMA={idx_extrema}, "
                      f"{self.data_str}_fit"
                      f"={f_fit.format(*p)}, "
                      f"K={K:5.3f}, idx_ref={idx_ref}")

            # decide whether we are going to stop iterating:
            terminate = False
            if len(idx_extrema) < 2*N+1:  # didn't find the desired number of extrema
                terminate = True
            elif idx_extrema[N] < idx_ref:
                # The N+1-st extremum should be *later* than idx_ref.  If not, sth wrong
                terminate = True

            if len(idx_extrema) >= 2*N-1:
                # at most two extrema short of target.  Assume the fit is
                # good enough to report extrema obtained through it.
                if count == 1:
                    # at first call (at start of waveform, report the extrema
                    # identified in the left part of the fitting interval
                    for k in range(0, N):
                        extrema.append(idx_extrema[k])
                        t_extrema_refined.append(extrema_refined[0][k])
                        data_extrema_refined.append(extrema_refined[1][k])
                        phase_gw_extrema_refined.append(extrema_refined[2][k])
                if idx_extrema[N] > idx_ref:
                    # the N+1-st extremum should be *after* idx_ref.  If it is
                    # not, something went wrong, so do not take this extremum

                    # take the extremum in the middle of the fitting interval.
                    # (if we are short extrema, then there will be fewer to the
                    # right. To not report those, due to potentially inaccurate
                    # fits that close to merger)
                    extrema.append(idx_extrema[N])
                    t_extrema_refined.append(extrema_refined[0][N])
                    data_extrema_refined.append(extrema_refined[1][N])
                    phase_gw_extrema_refined.append(extrema_refined[2][N])

                    # also store additional extrema, just in case we want to
                    # return them
                    extra_extrema = idx_extrema[N+1:]
                    extra_extrema_t = extrema_refined[0][N+1:]
                    extra_extrema_data = extrema_refined[1][N+1:]
                    extra_extrema_phase_gw = extrema_refined[2][N+1:]

                if terminate:
                    # be greedy and take any further extrema we know
                    for k in range(N+1, len(idx_extrema)):
                        if idx_extrema[k] > idx_ref:
                            extrema.append(idx_extrema[k])
                            t_extrema_refined.append(extrema_refined[0][k])
                            data_extrema_refined.append(
                                extrema_refined[1][k])
                            phase_gw_extrema_refined.append(
                                extrema_refined[2][k])
            else:
                # more than two extrema short.  In this case, don't trust the
                # fit anymore, but rather report the extema found in the last
                # iteration

                # sanity check: if we have too few extrema, ten terminate must
                # be set, too.
                if not terminate:
                    raise Exception("Logical error -- should never get here")
                if self.verbose:
                    print("terminating with very few extrema in this "
                          "iteration. Take left-over extrema from previous "
                          "iteration")
                    print(f"extra_extrema_t={extra_extrema_t}")
                extrema.extend(extra_extrema)
                t_extrema_refined.extend(extra_extrema_t)
                data_extrema_refined.extend(extra_extrema_data)
                phase_gw_extrema_refined.extend(extra_extrema_phase_gw)

            if terminate:
                # print("WARNING - TOO FEW EXTREMA FOUND. THIS IS LIKELY SIGNAL
                # THAT WE ARE AT MERGER")
                break

            # shift idx_ref one extremum to the right, in preparation for the
            # next extrema-search
            idx_ref = int(0.5*(idx_extrema[N]+idx_extrema[N+1]))

            if count > 10000:
                if pp:
                    pp.close()
                raise Exception("Detected more than 10000 extrema. "
                                "This has triggered a saftey exception."
                                "If your waveform is really this long, "
                                "please remove this exception and run again.")
        if self.verbose:
            print(f"Reached end of data.  Identified extrema = {extrema}")
        if pp:
            pp.close()

        if sign > 0:
            self.periastron_info = [np.array(t_extrema_refined),
                                    np.array(data_extrema_refined),
                                    np.array(phase_gw_extrema_refined)]
        else:
            self.apastron_info = [np.array(t_extrema_refined),
                                  np.array(data_extrema_refined),
                                  np.array(phase_gw_extrema_refined)]

        # Now that we are done with finding peaks, we shift the time axis to
        # it's original values
        self.t += self.t_merger
        return np.array(extrema)

        # Procedure:
        # - find length of useable dataset
        #     - exlude from start
        #     - find Max(A) and exclude from end
        # - one global fit for initial fitting parameters
        # - check which trefs we can do:
        #     - delineate N_extrema * 0.5 orbits from start
        #     - delineate N_extrema * 0.5 orbits from end
        #     - discard tref outside this interval (this places the trefs at
        #       least mostly into the middle of the fitting intervals. Not
        #       perfectly, since due to periastron advance the radial periods
        #       are longer than the orbital ones)
        # - set K=1
        # - set fitting_func=global_fit
        # - Loop over tref:
        #     - set old_extrema = [0.]*N_extrema
        #     - Loop over fitting-iterations:
        #         - (A) find interval that covers phase from
        #           K*(0.5*N_extrema+0.2) orbits before to after t_ref
        #         - find extrema of data - fit
        #         - update K based on the identified extrema
        #         - if  number of extrema != N_extrema:
        #             goto (A)  [i.e. compute a larger/smaller data-interval
        #             with new K]
        #         - if |extrema - old_extrema| < tol:  break
        #         - old_extrema=extrema
        #         - update fitting_func by fit to extrema

    def FindExtremaNearIdxRef(self,
                              idx_ref,
                              sign, Nbefore, Nafter, K,
                              f_fit, p_initial, bounds,
                              TOL,
                              increase_idx_ref_if_needed=True,
                              refine_extrema=False,
                              verbose=False,
                              pp=None,
                              plot_info=""):
        """given a GW modes data (t, phase_gw, data), identify a stretch of
        data [idx_lo, idx_hi] centered roughly around the index idx_ref which
        satisfies the following properties:

          - The interval [idx_lo, idx_hi] contains Nbefore+Nafter maxima
            (if sign==+1) or minimia (if sign==-1)
            of trend-subtracted data, where Nbefore exrema are before idx_ref
            and Nafter extrema are after idx_ref
          - The trend-subtraction is specified by the fitting function
            `data_trend = f_fit(t, *p)`.
            Its fitting parameters `*p` are self-consistently fitted to the
            N_extrema extrema.
          - if increase_idx_ref_if_needed, idx_ref is allowed to increase in
            order to reach the desired Nbefore.

        ASSUMPTIONS & POSSIBLE FAILURE MODES

          - if increase_idx_ref_if_needed == False, and idx_lo cannot be
            reduced enough to reach Nbefore -> raise Exception
          - if fewer extrema are identified than requested, then the function
            will return normally, but with len(idx_extrema) **SMALLER** than
            Nbefore+Nafter. This signals that the end of the data is reached,
            and that the user should not press to even larger idx_ref.

        Parameters
        ----------
        idx_ref
            the reference index, i.e. the approximate middle of the
            interval of data to be sought
        sign
            if +1, look for maxima, if -1, look for minima
        Nbefore
            Number of extrema to identify before idx_ref
        Nafter
            Number of extrema to identify after idx_ref
            if Nafter=Nbefore-1, then the Nbefore'th extremum
            will be centered
        K
            an estimate for the periastron advance of the binary,
            i.e. the increase of phase_gw/4pi between two extrema
        f_fit
            fitting function :func:`f_fit(t, *p)` to use for trend-subtraction
        p_initial
            initial guesses for the best-fit parametes
        p_bounds
            bounds for the fit-parameters
        TOL
            Iterate until the maximum change in any one data at an
            extremum is less tha this TOL
        increase_idx_ref_if_needed
            if true, allows to increase idx_ref in order to achieve Nbefore
            extrema between start of dataset and idx_ref (idx_ref will never be
            decreased, in order to preserve monotonicity to help tracing out an
            inspiral)

        pp
            a PdfPages object for a diagnostic output plot
        plot_info
            string placed into the title of the diagnostic plot

        Returns
        -------
        idx_extrema
            the indices of the identified extrema USUALLY
            len(idx_extrema) == Nbefore+Nafter HOWEVER,
            if not enough extrema can be identified
            (e.g.  end of data), then a shorter or even empty list can be
            returned

        p
            the fitting parameters of the best fit through the extrema
        K
            an updated estimate of the periastron advance K (i.e. the average
            increase of phase_gw between extrema divided by 4pi)
        idx_ref
            a (potentially increased) value of idx_ref, so that Nbefore extrema
            were found between the start of the data and idx_ref
        extrema_refined
            (t_extrema_refined, data_extrema_refined, phase_gw_extrema_refined)
            information about the
            parabolic-fit-refined extrema.  If RefineExtrema==True, these
            arrays have same length as idx_extrema.  Otherwise, empty.
        """
        extrema_type = {+1: "pericenters", -1: "apocenters"}[sign]
        if verbose:
            print(f"FindExtremaNearIdxRef  idx_ref={idx_ref}, "
                  f"K_initial={K:5.3f}, "
                  f"p_initial={f_fit.format(*p_initial)}"
                  f", refine_extrema={refine_extrema}")

        # look for somewhat more data than we (probably) need
        DeltaPhase = 4.2*np.pi*K
        idx_lo = np.argmax(
            self.phase_gw > self.phase_gw[idx_ref]
            - DeltaPhase*Nbefore)
        idx_hi = np.argmax(
            self.phase_gw > self.phase_gw[idx_ref]
            + DeltaPhase*Nafter)
        if idx_hi == 0:
            idx_hi = len(self.phase_gw)
            if verbose:
                print("WARNING: reaching end of data, so close to merger")
        p = p_initial
        it = 0

        old_extrema = np.zeros(Nbefore+Nafter)
        old_idx_lo, old_idx_hi = -1, -1

        # this variable counts the number of iterations in which Nright was one
        # too few.  This is used to detect limiting cycles, where the interval
        # adjustment oscillates between
        #   short interval with Nleft, Nright-1    extrema
        #   long interval with  Nleft, Nright      extrema
        # the oscillations can occur, because the fit with one more/left
        # extremum is so different as to make the extremum appear/vanish
        Count_Nright_short = 0

        if pp:
            fig, axs = plt.subplots(1, 3, figsize=(12, 4))
            fig.suptitle(plot_info, x=0.02, ha='left')
            axs[0].set_title(
                'trend-subtracted:  '
                f"{'-' if sign == -1 else ''}"
                f"({self.label_for_data_for_finding_extrema}-"
                f'{self.label_for_fit_to_data_for_finding_extrema})',
                fontsize='small')
            axs[1].set_title(f'{self.label_for_data_for_finding_extrema}(t)')
            axs[2].set_title('residual of fit')
            plot_offset = None

        # compute prominence
        distance, prominence = self.compute_distance_and_prominence(
            idx_lo, idx_hi, f_fit, p)
        if verbose:
            print(f"       find_peaks: distance={distance}, "
                  f"prominence={prominence}")
        interval_changed_on_it = -1
        if self.return_diagnostic_data:
            # Append an empty dict and for each iteration add data to this dict with iteration
            # number as the key.
            self.diagnostic_data_dict['params'][extrema_type].append({})
            self.diagnostic_data_dict['t_extrema'][extrema_type].append({})
            self.diagnostic_data_dict['data_extrema'][extrema_type].append({})
            self.diagnostic_data_dict['t_ref'][ extrema_type].append({})
            self.diagnostic_data_dict['f_fit'][ extrema_type].append({})
        while True:
            it = it+1
            if verbose:
                print(f"it={it}:  [{idx_lo} / {idx_ref} / {idx_hi}],  "
                      f"K={K:5.3f}")
            data_residual = (self.data_for_finding_extrema[idx_lo:idx_hi]
                             - f_fit(self.t[idx_lo:idx_hi], *p))
            data_residual_amp = max(data_residual)-min(data_residual)
            # TODO -- pass user-specified arguments into find_peaks
            # POSSIBLE UPGRADE
            # find_peaks on discrete data will not identify a peak to a
            # location better than the time-spacing.  To improve, one can add a
            # parabolic fit to the data around the time of extremum, and then
            # take the (fractional) index where the fit reaches its maximum.
            # Harald has some code to do this, but he hasn't moved it over yet
            # to keep the base implementation simple.

            # width used as exclusion in find_peaks
            #    1/2 phi-orbit  (at highest data)
            #    translated into samples using the maximum time-spacing

            idx_extrema, properties = scipy.signal.find_peaks(
                sign*data_residual,
                distance=distance,
                prominence=prominence)

            # add offset due to to calling find_peaks with sliced data
            idx_extrema = idx_extrema+idx_lo
            Nleft = sum(idx_extrema < idx_ref)
            Nright = sum(idx_extrema >= idx_ref)

            # remember info about extrema to be used in rest of this function
            N_extrema = len(idx_extrema)
            t_extrema = self.t[idx_extrema]
            data_extrema = self.data_for_finding_extrema[idx_extrema]
            phase_gw_extrema = self.phase_gw[idx_extrema]
            # data_residual is shorter array
            data_residual_extrema = data_residual[idx_extrema-idx_lo]
            # update K based on identified peaks
            if N_extrema >= 2:
                K = ((phase_gw_extrema[-1] - phase_gw_extrema[0])
                     / (4*np.pi * (N_extrema - 1)))
            if verbose:
                with np.printoptions(precision=2):
                    print(f"       idx_extrema=   {idx_extrema}, "
                          f"Nleft={Nleft},"
                          f" Nright={Nright}")
                    print(f"       t[idx_extrema]={t_extrema}")

            if refine_extrema:
                t_extrema, data_extrema, phase_gw_extrema \
                    = self.get_refined_extrema(t_extrema,
                                               data_extrema,
                                               phase_gw_extrema,
                                               verbose, f_fit, p)
                if verbose:
                    with np.printoptions(precision=4):
                        print("")
                        print("       Delta t_extrema = "
                              f"{t_extrema - self.t[idx_extrema]}")
            if N_extrema > 0 and self.return_diagnostic_data:
                # Update the dictionary for current extrema count with data for this iteration
                self.diagnostic_data_dict['params'][extrema_type][-1].update({it: p})
                self.diagnostic_data_dict['t_extrema'][extrema_type][-1].update({it: t_extrema})
                self.diagnostic_data_dict['data_extrema'][extrema_type][-1].update({it: data_extrema})
                self.diagnostic_data_dict['t_ref'][ extrema_type][-1].update({it: self.t[idx_ref]})
                self.diagnostic_data_dict['f_fit'][ extrema_type][-1].update({it: f_fit})
            if pp:
                # offset data vertically by 10^k*it
                if plot_offset is None:
                    plot_offset = 10**np.ceil(np.log10(data_residual_amp/2.))
                    axs[0].axvline(self.t[idx_ref], linestyle='--',
                                   color='grey', linewidth=1)

                line, = axs[0].plot(self.t[idx_lo:idx_hi],
                                    it*plot_offset
                                    + sign*data_residual, label=f"it={it}")
                if N_extrema > 0:
                    axs[0].plot(t_extrema,
                                it*plot_offset+sign*data_residual_extrema,
                                'o',
                                color=line.get_color())
                line, = axs[1].plot(
                    self.t[idx_lo:idx_hi],
                    self.data_for_finding_extrema[idx_lo:idx_hi]
                    + plot_offset*it)
                # note: 'line.get_color()` is also used below in axs[2].plot
                axs[1].plot(t_extrema, data_extrema+plot_offset*it, 'o',
                            color=line.get_color(), label=f"it={it}")

            if Nright < Nafter:  # and Nleft==Nbefore:
                Count_Nright_short = Count_Nright_short+1
                if verbose:
                    print(f"       Count_Nright_short={Count_Nright_short}")

            if N_extrema == 0 or it > 20:
                if verbose:
                    if N_extrema == 0:
                        print("could not identify a single extremum. "
                              "This can happen, for instance\n"
                              "for low eccentricity late in the inspiral where"
                              " the range of data\n"
                              "is so large that the prominence = "
                              f"{0.03*data_residual_amp} cannot be\n"
                              "reached by the small eccentricity oscillations")
                    else:
                        print(f"interval finding failed after it={it} "
                              "iterations. Exit")
                if pp:
                    # plt.legend()
                    fig.tight_layout()
                    fig.savefig(pp, format='pdf')
                    plt.close(fig)
                # don't really know what to do if we didn't identify any
                # extrema.  so return with empty idx_extrema and let upstream
                # code handle this
                return idx_extrema, p, K, idx_ref, [t_extrema, data_extrema,
                                                    phase_gw_extrema]

            if Nleft != Nbefore or Nright != Nafter:
                # number of extrema not as we wished, so update [idx_lo, idx_hi]
                # too many peaks left, discard by placing idx_lo between N and
                # N+1's peak to left
                if Nleft > Nbefore:
                    idx_lo = int(
                        (idx_extrema[Nleft-Nbefore-1]
                         + idx_extrema[Nleft-Nbefore])/2)
                    if verbose:
                        print(f"       idx_lo increased to {idx_lo}")
                elif Nleft < Nbefore:  # reduce idx_lo to capture one more peak
                    if idx_lo == 0:
                        # no more data to the left, so consider shifting
                        # idx_ref
                        if increase_idx_ref_if_needed:
                            if Nright >= 2:
                                # we need at least two maxima to the right to
                                # average for the new idx_ref
                                tmp = np.argmax(idx_extrema >= idx_ref)
                                # shift idx_ref one extremum to the right
                                idx_ref = int((idx_extrema[tmp]
                                               + idx_extrema[tmp + 1])/2)
                                # reflect the change in idx_ref to aid in
                                # updating idx_hi
                                Nright = Nright - 1
                                if verbose:
                                    print("       idx_ref increased to"
                                          f" {idx_ref}")
                            else:
                                pass
                                # First, wait for the idx_hi-updating below to
                                # widen the interval.  The next iteration will
                                # come back here and update idx_ref

                        else:
                            raise Exception(
                                f"could not identify {Nbefore}"
                                f" extrema to the left of idx_ref={idx_ref}")
                    else:
                        # decrease idx_lo by 0.6 radial periods.  This should
                        # get idx_lo toward seeing one earlier extremum.
                        # Rationale for 0.6: The next extremum should be 1
                        # radial period earlier.  We rather prefer to err on
                        # the low side, than overshooting and adding two
                        # extrema at once.
                        phase_lo = self.phase_gw[idx_lo] - K*4*np.pi*0.6
                        idx_lo = np.argmax(self.phase_gw > phase_lo)
                        if verbose:
                            print(f"       idx_lo reduced to {idx_lo}")
                # too many peaks to the right right, discard by placing idx_hi
                # between N and N+1's peak to right
                if Nright > Nafter:
                    idx_hi = int((idx_extrema[Nafter-Nright]
                                  + idx_extrema[Nafter-Nright-1])/2)
                    if verbose:
                        print(f"        idx_hi reduced to {idx_hi}")
                elif Nright < Nafter:
                    # increase idx_hi to capture one more peak
                    # do we have extra data?
                    if idx_hi < len(self.phase_gw):
                        # target phase on right 0.6 radial periods beyond
                        # current end of interval rationale for 0.6: The next
                        # extremum should be 1 radial period away, we are
                        # worried that near the end of the run, this prediction
                        # may not be accurate.  Therefore, go more slowly.
                        phase_hi = self.phase_gw[idx_hi] + K*4*np.pi*0.6
                        idx_hi = np.argmax(self.phase_gw > phase_hi)
                        if idx_hi == 0:
                            # coulnd't get as much data as we wished, take all
                            # we have
                            idx_hi = len(self.phase_gw)
                        if verbose and idx_hi != old_idx_hi:
                            print(f"       idx_hi increased to {idx_hi}")
                    else:
                        # we had already fully extended idx_hi in earlier
                        # iteration
                        if verbose:
                            print("        idx_hi at its maximum, "
                                  "but still insufficient "
                                  f"Nright={Nright}")

                if (idx_lo, idx_hi) != (old_idx_lo, old_idx_hi):
                    interval_changed_on_it = it
                    # remember when we last changed the search interval
                    (old_idx_lo, old_idx_hi) = (idx_lo, idx_hi)
                    # data-interval was updated; go back to start of loop to
                    # re-identify extrema
                    continue

            # if the code gets here, we have an interval [idx_lo,idx_high] with
            # either
            #  - Nleft + Nright envelope-subtracted extrema,
            # *or*
            #  - fewer envelope subtracted extrema and idx_hi at the end of the
            #  - data
            #
            # The following arrays are filled with information at the extrema
            #   t_extrema, phase_gw_extrema, data_extrema, where:
            #      * If refine_extrema==False: the arrays correspond to index
            #      * positions idx_extrema If refine_extrema==True: the arrays
            #      * are refined via fits.
            #
            # Now check whether data-envelope fitting has already converged.
            # If yes: return
            # If no:  re-fit envelope

            if N_extrema != len(old_extrema):
                # number of extrema has changed since last iteration, so avoid
                # to compute differences to last-iteration's extrema
                max_delta_data = 1e99
            else:
                max_delta_data = max(np.abs(data_extrema-old_extrema))
            if it >= 16 \
               and Nleft == Nbefore and Nright == Nafter \
               and interval_changed_on_it == it-1:
                # we have been interating for a while, and the recent change in
                # N_extrema points toward a limiting cycle.  Since we just
                # happened to have hit the right number of extrema, let's take
                # them and exit.
                if verbose:
                    print("looks like we hit a limiting cycle; presently the "
                          "number of extrema is correct, so exit")
                if pp:
                    # plt.legend()
                    fig.tight_layout()
                    fig.savefig(pp, format='pdf')
                    plt.close(fig)
                return idx_extrema, p, K, idx_ref, [t_extrema, data_extrema,
                                                    phase_gw_extrema]

            if Count_Nright_short >= 5 or N_extrema < 5 or it > 20:
                # safety exit to catch periodic loops note that
                # Count_Nright_short is only increased if Nright<Nafter,
                # therefore, this will coincide with Nright<Nafter, also
                # signaling that the overall extrema searching is ending.  we
                # require **5** extrema, in order to have safety for the **3**
                # parameter fit below
                if verbose:
                    print("exiting because Count_right_short="
                          f"{Count_Nright_short}"
                          f" is large, or N_extrema={N_extrema} is "
                          "insufficient")
                if pp:
                    # plt.legend()
                    fig.tight_layout()
                    fig.savefig(pp, format='pdf')
                    plt.close(fig)
                return idx_extrema, p, K, idx_ref, [t_extrema, data_extrema,
                                                    phase_gw_extrema]

            if max_delta_data < TOL:
                # (this cannot trigger on first iteration, due to
                # initialization of old_extrema)
                if verbose:
                    print(f"max_delta_{self.data_str}={max_delta_data:5.4g}<"
                          f"TOL={TOL}. Done.")
                if pp:
                    # plt.legend()
                    fig.tight_layout()
                    fig.savefig(pp, format='pdf')
                    plt.close(fig)
                return idx_extrema, p, K, idx_ref, [t_extrema, data_extrema,
                                                    phase_gw_extrema]
            p, pconv = scipy.optimize.curve_fit(f_fit, t_extrema,
                                                data_extrema, p0=p,
                                                bounds=bounds, maxfev=10000)
            if verbose and False:
                print("    PRODUCTION FIT: residual="
                      f"{np.linalg.norm(f_fit(t_extrema,*p)-data_extrema)},"
                      f" f={f_fit.format(*p)}")
            if pp:
                axs[2].plot(t_extrema, f_fit(t_extrema, *p)-data_extrema, "o",
                            color=line.get_color())
            old_extrema = data_extrema
            if verbose:
                print(f"       max_delta_{self.data_str}={max_delta_data:5.4g}"
                      f" => fit updated to f_fit={f_fit.format(*p)}")
        raise Exception("Should never get here")

    def compute_distance_and_prominence(self, idx_lo, idx_hi, f_fit, p):
        """Compute distance and prominence for current data section."""
        maxdt = np.max(np.diff(self.t[idx_lo:idx_hi]))

        # average orbital period during [idx_lo, idx_hi] idx_hi-1 also
        # works in the case when idx_hi = one-past-last-element
        T_orbit = ((self.t[idx_hi-1] - self.t[idx_lo])
                   / (self.phase_gw[idx_hi-1]
                      - self.phase_gw[idx_lo])
                   * 4*np.pi)

        # set distance = distance_factor * period.  This should exclude
        # spurious extrema due to noise
        distance = int(self.distance_factor*T_orbit/maxdt)
        data_residual = (self.data_for_finding_extrema[idx_lo:idx_hi]
                         - f_fit(self.t[idx_lo:idx_hi], *p))
        data_residual_amp = max(data_residual)-min(data_residual)
        prominence = data_residual_amp*self.prominence_factor

        return distance, prominence

    def get_refined_extrema(self, t_extrema, data_extrema, phase_gw_extrema,
                            verbose, f_fit, p):
        """Get refined extrema.

        Perform parabolic fits to data_residual around each extremum
        in order to find the time of extrema to sub-index
        accuracy.
        """
        if verbose:
            print(f"       Refine {len(t_extrema)} extremas, "
                  f"local fit with "
                  "Npoints=", end='')
        for k in range(len(t_extrema)):
            # length of fitting interval = 0.05radians left/right
            deltaT = 0.05 / data_extrema[k]
            idx_refine = np.abs(self.t - t_extrema[k]) < deltaT
            # number of points to be used in fit
            N_refine = sum(idx_refine)
            if verbose:
                print(f"{N_refine}  ", end='')
            if N_refine >= 7:  # enough data for fit
                t_parafit = self.t[idx_refine]
                # re-compute fit-subtracted data_residual, to avoid
                # indexing problems, should idx_lo/idx_high be so close
                # that the parabolic fitting interval extends beyond it
                data_resi_parafit = (
                    self.data_for_finding_extrema[idx_refine]
                    - f_fit(t_parafit, *p))

                parabola = np.polynomial.polynomial.Polynomial.fit(
                    t_parafit, data_resi_parafit, 2)
                t_max = parabola.deriv().roots()[0]

                # update extrema information
                t_extrema[k] = t_max

                # interpolate data from fits *assumption* the
                # fitting-interval is short enough that this is
                # accurate
                data_extrema[k] = parabola(t_max) + f_fit(t_max, *p)

                # 3rd order fit to phase to interpolate *assumption*
                # the fitting-interval is short enough that this is
                # accurate
                phase_fit = np.polynomial.polynomial.Polynomial.fit(
                    t_parafit, self.phase_gw[idx_refine], 3)
                phase_gw_extrema[k] = phase_fit(t_max)
            else:
                pass
                # if verbose:
                #    print(f"refinement of k={k} has too few points - skip")
        return t_extrema, data_extrema, phase_gw_extrema



"""Test smoothness of eccentricity definition methods using EMRI waveforms."""

__doc__ = """This test checks whether different eccentricity definitions vary
smoothly as a function of the internal eccentricity definition used by
Maarten's EMRI model. The EMRI waveforms that are used for this test are
generated from a fixed mass ratio, spins, and a fixed Momega0 (the initial
dimless orbital frequency) with eccentricity varying from 0 to 0.5. We try to
measure the eccentricity from these waveforms using different eccentricity
definitions. We test the smoothness of the measured eccentricities compared to
the EMRI eccentricities used by the waveform model to generate the waveforms in
the following two ways:

- For each waveform, we measure the eccentricity at the very first extrema
(pericenter or apocenter). That way, the measured eccentricity is also at
(nearly) Momega0. Finally, we plot the measured eccentricity vs the internal
EMRI eccentricity. You should check visually whether there are glitchy features
in these plots.

- For each waveform, We plot the measured eccentricity vs tref_out, and color
the lines by the input EMRI eccentricity at Momega0. You should check visually
whether there are glitchy features in these plots.

Usage: python emri_smoothness_test.py -d ecc_waveforms -m 'Amplitude' \
'ResidualAmplitude' -p 'all' python emri_smoothness_test.py -d ecc_waveforms -m
'all' -p 'all'
"""

import sys
import numpy as np
from tqdm import tqdm
import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.cm as cm
from mpl_toolkits.axes_grid1 import make_axes_locatable
import argparse
import pandas as pd
import glob
import re
sys.path.append("../../")
from gw_eccentricity import measure_eccentricity, get_available_methods
from gw_eccentricity.load_data import load_waveform
from gw_eccentricity.utils import SmartFormatter
from gw_eccentricity.plot_settings import use_fancy_plotsettings, figWidthsOneColDict
from gw_eccentricity.plot_settings import colorsDict, lwidths, lalphas, lstyles, labelsDict

parser = argparse.ArgumentParser(
    description=(__doc__),
    formatter_class=SmartFormatter)

parser.add_argument(
    "--data_dir", "-d",
    type=str,
    default="../../data/ecc_waveforms",
    help=("Base directory where waveform files are stored."))
parser.add_argument(
    "--method", "-m",
    type=str,
    nargs="+",
    default="all",
    help=("EccDefinition method to test. Can be 'all' OR one or more of the "
          f"methods in {get_available_methods()}."))
parser.add_argument(
    "--param_set_key", "-p",
    type=str,
    default="all",
    nargs="+",
    help=("Run test for this set of parameters kept fixed.\n"
          "Possible choices are 'all' OR one or more of 1, 2, 3.\n"
          "1: q=10, chi1z=chi2z=0.\n"
          "2: q=100, chi1z=chi2z=0.\n"
          "3: q=1000, chi1z=chi2z=0.\n"))
parser.add_argument(
    "--fig_dir",
    "-f",
    type=str,
    default='.',
    help="Directory to save figure.")
parser.add_argument(
    "--plot_format",
    type=str,
    default="png",
    help=("Format to save the plot. "
          "Can be any format that matplotlib supports."))
parser.add_argument(
    "--example",
    action="store_true",
    help=("This will override the figure name (that contains the "
          "information about parameter set, method used and so on)"
          " and uses a figure name which is of the form test_name_example.png"
          "where test_name is the name of the test."))
parser.add_argument(
    "--paper",
    action="store_true",
    help="Remove markers for paper.")
parser.add_argument(
    "--debug_index",
    type=int,
    default=None,
    help=("Only analyse the waveform with this index, "
          "and enable debugging output"))
parser.add_argument(
    "--verbose",
    action='store_true',
    default=False,
    help="increase verbosity")

args = parser.parse_args()

# check that given method is available
for method in args.method:
    if method not in get_available_methods() and args.method != "all":
        raise KeyError(f"Method {method} is not an allowed method."
                       f" Must be one of {get_available_methods()}"
                       " or `all` for using all available methods.")
# Format: [q, chi1z, chi2z]
available_param_sets = {
    "1": [10, 0, 0],
    "2": [100, 0, 0],
    "3": [1000, 0, 0]
}
# check that given param set is available
for p in args.param_set_key:
    if p not in available_param_sets and args.param_set_key != "all":
        raise KeyError(f"Param set key {p} is not an allowed param set key."
                       f" Must be one of {list(available_param_sets.keys())}"
                       " or `all` for using all available param set keys.")

data_dir = args.data_dir + "/Non-Precessing/EMRI/"

# set debug to False
# The default width is slightly larger than what we need.
# Setting it to smaller value helps finding all the extrema
extra_kwargs = {"debug": False,
                # "extrema_finding_kwargs": {"width": 300}
                }

if args.debug_index is not None:
    extra_kwargs['debug'] = True

cmap = cm.get_cmap("plasma")


def get_file_names(key):
    """Get the list of file paths based on key."""
    q, chi1z, chi2z = available_param_sets[key]
    fileNames = sorted(
        glob.glob(f"{data_dir}/q{q}/EMRI_0PA_q{q}_e*[!_ecc].h5"))
    return fileNames


def get_color(ecc, eccs, colors):
    """Get color for the ecc vs time lines."""
    return colors[np.argmin(np.abs(eccs - ecc))]


def plot_waveform_ecc_vs_model_ecc(methods, key):
    """Create plots for given methods and parameter set."""
    # Get the output figure name
    # method_str is used in the filename for the output figure
    if "all" in methods:
        methods = get_available_methods()
        method_str = "all"
    else:
        method_str = "_".join(methods)

    if args.example:
        fig_ecc_at_start_name = (
            f"{args.fig_dir}/test_emri_measured_ecc_vs_emri_ecc_example"
            f".{args.plot_format}")
        fig_ecc_vs_t_name = (
            f"{args.fig_dir}/test_emri_measured_ecc_vs_time_example."
            f"{args.plot_format}")
    else:
        fig_ecc_at_start_name = (
            f"{args.fig_dir}/test_emri_measured_ecc_vs_emri_ecc"
            f"_set{key}_{method_str}.{args.plot_format}")
        fig_ecc_vs_t_name = (
            f"{args.fig_dir}/test_emri_measured_ecc_vs_time_set{key}_"
            f"{method_str}.{args.plot_format}")
    # For emri vs measured ecc at the first extrema
    fig_ecc_at_start, ax_ecc_at_start = plt.subplots(
        figsize=(figWidthsOneColDict[style], 2 if args.paper else 3))
    # For measured ecc vs time
    fig_ecc_vs_t, ax_ecc_vs_t = plt.subplots(
        nrows=len(methods),
        figsize=(figWidthsOneColDict[style],  # Figure width
                 (2 if args.paper else 3)  # Height of each row
                 * len(methods)  # Number of rows
                 ))
    if len(methods) == 1:
        ax_ecc_vs_t = [ax_ecc_vs_t]
    # Create dictionaries to store different quantities for each methods
    # as we loop over all the eccentricity and methods.
    model_eccs = {}  # To keep track of eob input eccentricity.
    measured_eccs_at_start = {}  # To keep track of measured ecc at the first
    # extrema.
    tmaxList = {}  # To keep track of minimum time in tref_out across all eccs.
    tminList = {}  # To keep track of maximum time in tref_out across all eccs.
    failed_eccs = {}  # To keep track of failed cases.
    failed_indices = {}  # To Keep track of failed indices
    # Initiate the dictionary with an empty list for each method
    for method in methods:
        model_eccs.update({method: []})
        measured_eccs_at_start.update({method: []})
        tmaxList.update({method: []})
        tminList.update({method: []})
        failed_eccs.update({method: []})
        failed_indices.update({method: []})
    q, chi1z, chi2z = available_param_sets[key]
    emri_ecc_label = labelsDict['omega_start']
    # We exclude the first file since it is the quasicircular waveform
    filePaths = get_file_names(key)[1:]
    EMRIeccs = []
    for f in filePaths:
        ecc = float(re.findall("\d{1}\.\d{3}", f)[0])
        EMRIeccs.append(ecc)
    EMRIeccs = np.array(EMRIeccs)
    # Create an eccs array to map colors to eccentricities
    # for the ecc vs time line plots
    eccs_for_colors = np.linspace(
        np.round(EMRIeccs.min(), 2),
        np.round(EMRIeccs.max(), 2),
        len(filePaths))
    colors = cmap(np.linspace(0, 1, len(filePaths)))
    for idx0, f in tqdm(enumerate(filePaths), disable=args.verbose):
        # in debugging mode, skip all but debug_index:
        if args.debug_index is not None:
            if idx0 != args.debug_index:
                continue
        # Need to change the resolution of the data, otherwise the omega22
        # average are not monotonically increasing
        kwargs = {"filepath": f,
                  "deltaT": 0.05}
        if args.verbose:
            print(f"idx={idx0}, {f}")
        # check if any residual method is in methods. If yes then load
        # zeroecc waveforms also.
        if "ResidualAmplitude" in methods or "ResidualFrequency" in methods:
            kwargs.update({"include_zero_ecc": True})
        dataDict = load_waveform(origin="EMRI", **kwargs)
        tref_in = dataDict["t"]
        for idx, method in enumerate(methods):
            try:
                gwecc_dict = measure_eccentricity(
                    tref_in=tref_in,
                    dataDict=dataDict,
                    method=method,
                    extra_kwargs=extra_kwargs)
                tref_out = gwecc_dict["tref_out"]
                measured_ecc = gwecc_dict["eccentricity"]
                model_eccs[method] += [EMRIeccs[idx0]]
                # Get the measured eccentricity at the first available index.
                # This corresponds to the first extrema that occurs after the
                # initial time
                measured_eccs_at_start[method] += [measured_ecc[0]]
                tmaxList[method] += [tref_out[-1]]
                tminList[method] += [tref_out[0]]
                # Add measured ecc vs time plot for each method to
                # corresponding axis.
                ax_ecc_vs_t[idx].plot(
                    tref_out,
                    measured_ecc,
                    c=get_color(EMRIeccs[idx0], eccs_for_colors, colors))
                if idx == len(methods) - 1:
                    ax_ecc_vs_t[idx].set_xlabel(labelsDict["t_dimless"])
            except Exception as e:
                # collected failures
                failed_eccs[method] += [EMRIeccs[idx0]]
                failed_indices[method] += [idx0]
                if args.debug_index is None:
                    # If verbose, print the exception message
                    if args.verbose:
                        print(f"Exception raised with the message {e}")
                else:
                    # if debugging, print entire traceback
                    raise
    # Iterate over methods to plots measured ecc at the first extrema vs emri
    # eccs collected above looping over all EMRI eccs for each methods
    for idx, method in enumerate(methods):
        # Plot measured eccs at the first extrema vs emri eccs for different
        # methods
        ax_ecc_at_start.plot(
            model_eccs[method], measured_eccs_at_start[method], label=method,
            c=colorsDict.get(method, "C0"),
            ls=lstyles.get(method, "-"),
            lw=lwidths.get(method, 1),
            alpha=lalphas.get(method, 1),
            marker=".",
            # marker=None if args.paper else "."  # No marker for paper
        )
        # Customize the measured ecc at the first extrema vs time plots
        ax_ecc_vs_t[idx].grid()
        if len(tmaxList[method]) >= 1:
            tmin = max(tminList[method])  # Choose the shortest
            tmax = max(tmaxList[method])
            ymax = max(measured_eccs_at_start[method])
            ymin = min(EMRIeccs)
            ax_ecc_vs_t[idx].set_xlim(tmin, tmax)
            ax_ecc_vs_t[idx].set_ylim(ymin, ymax)
        ax_ecc_vs_t[idx].set_ylabel(labelsDict["eccentricity"])
        # ax_ecc_vs_t[idx].set_yscale("log")
        # Add yticks.
        # ax_ecc_vs_t[idx].set_yticks(10.0**np.arange(-6.0, 1.0, 2.0))
        # Add text indicating the method used
        ax_ecc_vs_t[idx].text(0.95, 0.95, method, ha="right", va="top",
                              transform=ax_ecc_vs_t[idx].transAxes)
        # Add colorbar
        norm = mpl.colors.Normalize(vmin=eccs_for_colors.min(),
                                    vmax=eccs_for_colors.max())
        divider = make_axes_locatable(ax_ecc_vs_t[idx])
        cax = divider.append_axes('right', size='3%', pad=0.1)
        sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
        cbar = fig_ecc_vs_t.colorbar(sm, cax=cax, orientation='vertical')
        # Set yticks on colorbar
        # cbar.ax.set_yticks(10**np.arange(-7.0, 1.0))
        cbar.ax.tick_params(labelsize=8)
        cbar.set_label(
            rf"{labelsDict['geodesic_eccentricity']} at "
            rf"{emri_ecc_label}",
            size=10)
        if idx == 0:
            ax_ecc_vs_t[idx].set_title(
                rf"{labelsDict['q']}={q:.1f}, "
                fr"{labelsDict['chi1z']}={chi1z:.1f}, "
                rf"{labelsDict['chi2z']}={chi2z:.1f}",
                ha="center")

    # Customize measured eccs at the first extrema vs eob eccs
    ax_ecc_at_start.set_title(
        rf"{labelsDict['q']}={q:.1f}, "
        fr"{labelsDict['chi1z']}={chi1z:.1f}, "
        rf"{labelsDict['chi2z']}={chi2z:.1f}",
        ha="center")
    ax_ecc_at_start.legend(frameon=True)
    # set ticks
    ticks = np.arange(0, EMRIeccs.max(), 0.1)
    ticklabels = [f"{tick:.1f}" for tick in ticks]
    ax_ecc_at_start.set_xticks(ticks)
    ax_ecc_at_start.set_xticklabels(ticklabels)
    ax_ecc_at_start.set_yticks(ticks)
    ax_ecc_at_start.set_yticklabels(ticklabels)
    ax_ecc_at_start.set_xlabel(rf"{labelsDict['geodesic_eccentricity']} at "
                               rf"{emri_ecc_label}")
    ax_ecc_at_start.set_ylabel(labelsDict["eccentricity"])
    # ax_ecc_at_start.set_ylim(top=EMRIeccs.max())
    # ax_ecc_at_start.set_xlim(EMRIeccs.min(), EMRIeccs.max())
    ax_ecc_at_start.grid()

    # Save figures
    fig_ecc_at_start.savefig(
        f"{fig_ecc_at_start_name}",
        bbox_inches="tight")
    fig_ecc_vs_t.savefig(
        f"{fig_ecc_vs_t_name}",
        bbox_inches="tight")
    return failed_eccs, failed_indices


def report_failures(failed_eccs, failed_indices, methods):
    """Report failed cases."""
    if "all" in methods:
        methods = get_available_methods()
    for method in methods:
        num_failures = len(failed_eccs[method])
        print(f"================{method}============================")
        if num_failures == 0:
            print("All cases passed!")
        else:
            print(f"{num_failures} {'case' if num_failures == 1 else 'cases'} "
                  "failed.")
            df = pd.DataFrame({
                "Case indices": failed_indices[method],
                "Eccentricity": failed_eccs[method]
            })
            print(df)


if "all" in args.param_set_key:
    args.param_set_key = list(available_param_sets.keys())

#  Use fancy colors and other settings
style = "APS" if args.paper else "Notebook"
use_fancy_plotsettings(style=style)

for key in args.param_set_key:
    failed_eccs, failed_indices = plot_waveform_ecc_vs_model_ecc(
        args.method, key)
    print(f"================Test report for set {key}===================")
    report_failures(failed_eccs, failed_indices, args.method)



import nbformat
from nbconvert.preprocessors import ExecutePreprocessor
import platform
from glob import glob
import subprocess

git_home = subprocess.check_output(
    ['git', 'rev-parse', '--show-toplevel'], text=True).strip('\n')

python_version = platform.python_version()
if python_version[:2] == '2.':
    python_version = 'python2'
elif python_version[:2] == '3.':
    python_version = 'python3'


def test_example_notebooks():
    """ Tests that all the example ipython notebooks of format
    gw_eccentricity/examples/*.ipynb are working. Since we expect these to be
    used by our users, it would be emabarassing if our own examples failed.
    """
    notebooks_list = glob(f'{git_home}/examples/*.ipynb')
    notebooks_list.sort()
    if len(notebooks_list) == 0:
        raise Exception("No notebooks found!")
    # Some of the notebooks need data files which are not pushed to github due
    # to large file size. We skip these notebooks from testing.
    no_test_list = [f"{git_home}/examples/load_waveform_demo.ipynb"]
    for notebook in notebooks_list:
        if notebook in no_test_list:
            # skip these notebooks, otherwise it will fail due to missing data
            # file.
            print(f"No test for {notebook}")
            continue
        print(f'testing {notebook}')
        with open(notebook) as f:
            nb = nbformat.read(f, as_version=4)

        ep = ExecutePreprocessor(timeout=None, kernel_name=python_version)
        ep.preprocess(nb, {'metadata': {'path': '.'}})



"""
Find pericenters and apocenters using frequency.

Part of Eccentricity Definition project.
Md Arif Shaikh, Mar 28, 2022
"""
from .eccDefinitionUsingAmplitude import eccDefinitionUsingAmplitude


class eccDefinitionUsingFrequency(eccDefinitionUsingAmplitude):
    """Measure eccentricity by finding extrema location from frequency."""

    def __init__(self, *args, **kwargs):
        """Init for eccDefinitionUsingFrequency class.

        parameters:
        ----------
        dataDict: Dictionary containing the waveform data.
        """
        super().__init__(*args, **kwargs)
        self.label_for_data_for_finding_extrema = self.get_label_for_plots("omega")
        self.method = "Frequency"

    def get_data_for_finding_extrema(self):
        """Get the data for extrema finding."""
        return self.omega_gw



"""
Generate EOB waveforms using Toni's SEOBNRv4EHM model.

This script is mostly for me (Arif) to generate EOB waveforms for the
test_eob_vs_measured_ecc.py script and is not important for most of the
users. Only script to run for test is test_eob_vs_measured_ecc.py.
"""
import os
import argparse
from argparse import RawTextHelpFormatter
import numpy as np
from tqdm import tqdm
import lal
import h5py
import seobnrv4ehm as seob

parser = argparse.ArgumentParser(formatter_class=RawTextHelpFormatter)
parser.add_argument(
    "--data_dir",
    type=str,
    required=True,
    help=("Base directory where waveform files are to be stored."))
parser.add_argument(
    "--set",
    type=str,
    default="all",
    nargs="+",
    help=("Generate waveforms for the set of parameters.\n"
          "Possible choices are one or more of 1, 2, 3, 4 or all.\n"
          "1: q=1, chi1z=chi2z=0.\n"
          "2: q=2, chi1z=chi2z=0.5\n"
          "3: q=4, chi1z=chi2z=-0.6\n"
          "4: q=6, chi1z=0.4, chi2z=-0.4.\n"
          "Default is 'all'.\n"
          "usage: --set 1 2"))
parser.add_argument(
    "--ecc",
    type=float,
    default=[-1],
    nargs="+",
    help=("Generate waveforms for the ecc values.\n"
          "Default is -1 which means it will generate `args.num` waveforms"
          " from `10**args.emin` to `10**args.emax`\n"
          "usage: --ecc 0.1 0.2"))
parser.add_argument(
    "--Momega0",
    type=float,
    default=0.01,
    help="Initial Momega0 to generate waveform. Default is 0.01.")
parser.add_argument(
    "--emin",
    type=float,
    default=-7.0,
    help="Minimum eccentricity in log10.")
parser.add_argument(
    "--emax",
    type=float,
    default=0.0,
    help="Maximum eccentricity in log10.")
parser.add_argument(
    "--num",
    type=int,
    default=150,
    help="Number of waveforms")

args = parser.parse_args()
if -1 in args.ecc:
    args.ecc = 10.0**np.linspace(args.emin, args.emax,
                                 args.num)

param_sets = {"1": [1, 0, 0],
              "2": [2, 0.5, 0.5],
              "3": [4, -0.6, -0.6],
              "4": [6, 0.4, -0.4]}

if "all" in args.set:
    args.set = param_sets
else:
    sets = {}
    for key in args.set:
        sets.update({key: param_sets[key]})
    args.set = sets

mean_ano = np.pi/2
M = 50
MT = M * lal.MTSUN_SI
freq_in = args.Momega0 / (MT * np.pi)
print(freq_in)
deltaTOverM = 1
dt = deltaTOverM * MT
print(1/dt)

data_dir = args.data_dir + "Non-Precessing/EOB/"

if not os.path.exists(data_dir):
    os.system(f"mkdir -p {data_dir}")

for key in args.set:
    q, chi1z, chi2z = param_sets[key]
    for ecc in tqdm(args.ecc):
        times, modes = seob.get_modes(
            q=q,
            chi1=chi1z,
            chi2=chi2z,
            M_fed=M,
            delta_t=dt,
            f_min=freq_in,
            eccentricity=ecc,
            eccentric_anomaly=mean_ano,
            EccIC=-1,
            physical_units=False)
        fileName = (f"{data_dir}/EccTest_q{q:.2f}_chi1z{chi1z:.2f}_"
                    f"chi2z{chi2z:.2f}_EOBecc{ecc:.10f}_"
                    f"Momega0{args.Momega0:.3f}_meanAno{mean_ano:.3f}.h5")
        f = h5py.File(fileName, "w")
        f["(2, 2)"] = modes[2, 2]
        f["t"] = times
        dset = f.create_dataset("params", ())
        dset.attrs["q"] = q
        dset.attrs["deltaTOverM"] = deltaTOverM
        dset.attrs["Momega0"] = args.Momega0
        dset.attrs["ecc"] = ecc
        dset.attrs["mean_ano"] = mean_ano
        dset.attrs["chi1z"] = chi1z
        dset.attrs["chi2z"] = chi2z
        f.close()

print(f"files are saved at {data_dir}")



"""Truncate waveform by flow."""
from gw_eccentricity import get_available_methods
import numpy as np
import copy


def truncate_waveform_by_flow(dataDict=None,
                              flow=None,
                              m_max=None,
                              method="Amplitude",
                              extra_kwargs=None):
    """Truncate waveform by flow.

    Eccentric waveforms have a non-monotonic instantaneous frequency.
    Therefore, truncating the waveform by demanding that the truncated waveform
    should contain all frequencies that are greater than or equal to a given
    minimum frequency, say flow, must be done carefully since the instantaneous
    frequency can be equal to the given flow at multiple points in time.

    We need to find the time tlow, such that all the frequencies at t < tlow
    are < flow and therefore the t >= tlow part of the waveform would
    retain all the frequencies that are >= flow. Note that the t >= tlow part
    could contain some frequencies < flow but that is fine, all we need is not
    to lose any frequencies >= flow.

    This could be done by using the frequency interpolant omega22_p(t) through
    the pericenters because
    1. It is a monotonic function of time.
    2. If at a time tlow, omega22_p(tlow) * (m_max/2) = 2*pi*flow, then all
    frequencies >= flow will be included in the waveform truncated at
    t=tlow. The m_max/2 factor ensures that this statement is true for all
    modes, as the frequency of the h_{l, m} mode scales approximately as m/2 *
    omega_22/(2*pi).

    Thus, we find tlow such that omega22_a(tlow) * (m_max/2) = 2*pi*flow and
    truncate the waveform by keeping only the part where t >= tlow.

    Paramerers:
    -----------
    datadict: dict
        Dictionary containing waveform data in the following format:
        dataDict = {"t": t,
                    "hlm": hlm},
        where t is the time array and hlm is a dictionary of waveform modes
        of the following format:
        hlm = {(l, m): lm_mode}
    flow: float
        Lower cutoff frequency to truncate the given waveform modes.
        The truncated waveform would have all the frequencies that are >= flow.
    m_max: int
        Maximum m (index of h_{l, m}) to account for while setting the tlow
        for truncation.  If None, then it is set using the highest available
        m from the modes in the dataDict.
        Default is None.
    method: str
        Method to find the locations of the apocenters. Default is "Amplitude".
        See gw_eccentricity.get_available_modes for available modes.
    extra_kwargs: dict
        Dictionary of arguments that might be used for extrema finding routine.
        Default values are set using eccDefinition.get_default_extra_kwargs.

    Returns:
    --------
    truncatedDict: dict
        Dictionary containing the truncated waveform.
        Has the same type as dataDict.
    gwecc_object: obj
        Object used to truncate the dataDict.
    """
    if dataDict is None:
        raise ValueError("dataDict can not be None.")
    available_methods = get_available_methods(return_dict=True)

    if method in available_methods:
        gwecc_object = available_methods[method](dataDict,
                                                 extra_kwargs=extra_kwargs)
        # Get the pericenters
        pericenters = gwecc_object.find_extrema("pericenters")
        original_pericenters = pericenters.copy()
        gwecc_object.check_num_extrema(pericenters, "pericenters")
        # Get the good pericenters
        pericenters = gwecc_object.drop_extrema_if_extrema_jumps(
            pericenters, 1.5, "pericenters")
        gwecc_object.pericenters_location = gwecc_object.drop_extrema_if_too_close(
            pericenters, extrema_type="pericenters")
        # Build the interpolants of omega22 at the extrema
        gwecc_object.omega22_pericenters_interp = gwecc_object.interp_extrema("pericenters")
    # If m_max is not provided, get the highest available m from the dataDict
    if m_max is None:
        modes = gwecc_object.dataDict["hlm"].keys()
        m_max = max([m for (l, m) in modes])
    # Find time where omega22_apocenter_interp(tlow) = 2 * pi * flow
    tmin = gwecc_object.t[gwecc_object.pericenters_location[0]]
    tmax = gwecc_object.t[gwecc_object.pericenters_location[-1]]
    gwecc_object.t_pericenters_interp = gwecc_object.t[
        np.logical_and(gwecc_object.t >= tmin,
                       gwecc_object.t <= tmax)]
    gwecc_object.f22_pericenters_interp \
        = gwecc_object.omega22_pericenters_interp(gwecc_object.t_pericenters_interp)/2/np.pi

    idx_low = np.where(
        gwecc_object.f22_pericenters_interp * (m_max/2) >= flow)[0][0]
    tlow = gwecc_object.t_pericenters_interp[idx_low]

    truncatedDict = copy.deepcopy(dataDict)
    for mode in truncatedDict["hlm"]:
        truncatedDict["hlm"][mode] \
            = truncatedDict["hlm"][mode][truncatedDict["t"] >= tlow]
    truncatedDict["t"] = truncatedDict["t"][truncatedDict["t"] >= tlow]

    gwecc_object.method = method
    gwecc_object.m_max = m_max
    gwecc_object.tlow_for_trucating = tlow
    gwecc_object.truncatedDict = truncatedDict
    gwecc_object.f_low_for_truncating = flow

    return truncatedDict, gwecc_object



"""Test smoothness of eccentricity definition methods using EOB waveforms."""

__doc__ = """This test checks whether different eccentricity definitions vary
smoothly as a function of the internal eccentricity definition used by Toni's
EOB model. The EOB waveforms that are used for this test are generated from a
fixed mass ratio, spins, and either a fixed Momega0 (the initial dimless
orbital frequency) or fixed length (varying the initial frequency), with
eccentricity varying from 1e-7 to 1 (for fixed Momega0) or to 0.89 (for fixed
length since ecc > 0.89 generates inaccurate waveforms of length 20000M). We
try to measure the eccentricity from these waveforms using different
eccentricity definitions. We test the smoothness of the measured eccentricities
compared to the eob eccentricities used by the waveform model to generate the
waveforms in the following two ways:

- For each waveform, we measure the eccentricity at the very first extrema
(pericenter or apocenter). That way, the measured eccentricity is also at
(nearly) Momega0 for waveforms starting with fixed Momega0 or at (neary)
-20000M for waveforms with fixed length. Finally, we plot the measured
eccentricity vs the internal EOB eccentricity.  You should check visually
whether there are glitchy features in these plots.

- For each waveform, We plot the measured eccentricity vs tref_out, and color
the lines by the input EOB eccentricity at Momega0. You should check visually
whether there are glitchy features in these plots.

Usage:
python eob_smoothness_test.py -d ecc_waveforms -m 'Amplitude' \
'ResidualAmplitude' -p 'all'
python eob_smoothness_test.py -d ecc_waveforms -m 'all' -p 'all'
"""

import sys
import numpy as np
from tqdm import tqdm
import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.cm as cm
from mpl_toolkits.axes_grid1 import make_axes_locatable
import argparse
import pandas as pd
import glob
sys.path.append("../../")
from gw_eccentricity import measure_eccentricity, get_available_methods
from gw_eccentricity.load_data import load_waveform
from gw_eccentricity.utils import SmartFormatter
from gw_eccentricity.plot_settings import use_fancy_plotsettings, figWidthsOneColDict
from gw_eccentricity.plot_settings import colorsDict, lwidths, lalphas, lstyles, labelsDict

parser = argparse.ArgumentParser(
    description=(__doc__),
    formatter_class=SmartFormatter)

parser.add_argument(
    "--data_dir", "-d",
    type=str,
    default="../../data/ecc_waveforms",
    help=("Base directory where waveform files are stored."))
parser.add_argument(
    "--method", "-m",
    type=str,
    nargs="+",
    default="all",
    help=("EccDefinition method to test. Can be 'all' OR one or more of the "
          f"methods in {get_available_methods()}."))
parser.add_argument(
    "--param_set_key", "-p",
    type=str,
    default="all",
    nargs="+",
    help=("Run test for this set of parameters kept fixed.\n"
          "Possible choices are 'all' OR one or more of 1, 2, 3, 4.\n"
          "1: q=1, chi1z=chi2z=0. With Momega0 = 0.01.\n"
          "2: q=2, chi1z=chi2z=0.5. With Momega0 = 0.01.\n"
          "3: q=4, chi1z=chi2z=-0.6.  With Momega0 = 0.01.\n"
          "4: q=6, chi1z=0.4, chi2z=-0.4. With Momega0 = 0.01.\n"
          "5: q=1, chi1z=chi2z=0. With fixed length of 20000M.\n"
          "6: q=4, chi1z=chi2z=-0.6. With fixed length of 20000M.\n"))
parser.add_argument(
    "--fig_dir",
    "-f",
    type=str,
    default='.',
    help="Directory to save figure.")
parser.add_argument(
    "--plot_format",
    type=str,
    default="png",
    help=("Format to save the plot. "
          "Can be any format that matplotlib supports."))
parser.add_argument(
    "--example",
    action="store_true",
    help=("This will override the figure name (that contains the "
          "information about parameter set, method used and so on)"
          " and uses a figure name which is of the form test_name_example.png"
          "where test_name is the name of the test."))
parser.add_argument(
    "--slice",
    type=int,
    default=1,
    help=("Slice the eccentricities to loop over (EOBecccs array containing"
          " 150 eccs from 1e-7 to 1.0) by taking only every nth ecc value"
          " given by slice. This is useful when we do not want to plot"
          " for all the ecc but skip n number of ecc given by slice."
          " NOTE: This is applied only for the measured ecc vs time plot to"
          " make each line readable. Specially for presenting in the paper."
          " For regular test using the default value, i.e.,"
          " no slicing is recommended."))
parser.add_argument(
    "--paper",
    action="store_true",
    help="Remove markers for paper.")
parser.add_argument(
    "--debug_index",
    type=int,
    default=None,
    help=("Only analyse the waveform with this index, "
          "and enable debugging output"))
parser.add_argument(
    "--verbose",
    action='store_true',
    default=False,
    help="increase verbosity")

args = parser.parse_args()

# check that given method is available
for method in args.method:
    if method not in get_available_methods() and args.method != "all":
        raise KeyError(f"Method {method} is not an allowed method."
                       f" Must be one of {get_available_methods()}"
                       " or `all` for using all available methods.")
# Format: [q, chi1z, chi2z]
available_param_sets = {
    "1": [1, 0, 0],        # With fixed Momega0 = 0.01
    "2": [2, 0.5, 0.5],    # With fixed Momega0 = 0.01
    "3": [4, -0.6, -0.6],  # With fixed Momega0 = 0.01
    "4": [6, 0.4, -0.4],   # With fixed Momega0 = 0.01
    "5": [1, 0, 0],        # With fixed length = 20000M
    "6": [4, -0.6, -0.6]   # With fixed length = 20000M
}
# check that given param set is available
for p in args.param_set_key:
    if p not in available_param_sets and args.param_set_key != "all":
        raise KeyError(f"Param set key {p} is not an allowed param set key."
                       f" Must be one of {list(available_param_sets.keys())}"
                       " or `all` for using all available param set keys.")

EOBeccs = 10**np.linspace(-7, 0, 150)
Momega0 = 0.01
# We generate the waveforms with initial mean anomaly = pi/2 so that over the
# entire range of initial eccentricities 1e-7 to 1, the length of generated
# waveforms (inspiral time) does not change too much which is the case if we
# start with mean anomaly = 0. This is because at mean_ano = pi/2 the
# instantaneous frequency is close to the orbit averaged frequency.
meanAno = np.pi/2
Momega0_zeroecc = 0.002
data_dir = args.data_dir + "/Non-Precessing/EOB/"

# set debug to False
extra_kwargs = {"debug": False}

if args.debug_index is not None:
    extra_kwargs['debug'] = True

cmap = cm.get_cmap("plasma")


def get_file_name(key, ecc):
    """Get file name based on key and given ecc."""
    q, chi1z, chi2z = available_param_sets[key]
    keys = list(available_param_sets.keys())
    if key in keys[:4]:
        fileName = (f"{data_dir}/EccTest_q{q:.2f}_chi1z{chi1z:.2f}_"
                    f"chi2z{chi2z:.2f}_EOBecc{ecc:.10f}_"
                    f"Momega0{Momega0:.3f}_meanAno{meanAno:.3f}.h5")
    if key in keys[4:]:
        fileName = glob.glob(
            f"{data_dir}/EccTest*_q{q:.2f}_chi1z{chi1z:.2f}_"
            f"chi2z{chi2z:.2f}_EOBecc{ecc:.10f}*"
            f"meanAno{meanAno:.3f}_20000.h5")[0]
    fileName_zeroecc = (
        f"{data_dir}/EccTest_q{q:.2f}_chi1z{chi1z:.2f}_"
        f"chi2z{chi2z:.2f}_EOBecc{0.0:.10f}_"
        f"Momega0{Momega0_zeroecc:.3f}_meanAno{meanAno:.3f}.h5")
    return fileName, fileName_zeroecc


def plot_waveform_ecc_vs_model_ecc(methods, key):
    """Create plots for given methods and parameter set."""
    # Get the output figure name
    # method_str is used in the filename for the output figure
    if "all" in methods:
        methods = get_available_methods()
        method_str = "all"
    else:
        method_str = "_".join(methods)

    if args.example:
        fig_ecc_at_start_name = (
            f"{args.fig_dir}/test_eob_vs_measured_ecc_example"
            f".{args.plot_format}")
        fig_ecc_vs_t_name = (
            f"{args.fig_dir}/test_measured_ecc_vs_time_example."
            f"{args.plot_format}")
    else:
        fig_ecc_at_start_name = (
            f"{args.fig_dir}/test_eob_vs_measured_ecc_set{key}_{method_str}"
            f".{args.plot_format}")
        fig_ecc_vs_t_name = (
            f"{args.fig_dir}/test_measured_ecc_vs_time_set{key}_{method_str}"
            f".{args.plot_format}")
    # For eob vs measured ecc at the first extrema
    fig_ecc_at_start, ax_ecc_at_start = plt.subplots(
        figsize=(figWidthsOneColDict[style], 2 if args.paper else 3))
    # For measured ecc vs time
    fig_ecc_vs_t, ax_ecc_vs_t = plt.subplots(
        nrows=len(methods),
        figsize=(figWidthsOneColDict[style],  # Figure width
                 (2 if args.paper else 3)  # Height of each row
                 * len(methods)  # Number of rows
                 ))
    if len(methods) == 1:
        ax_ecc_vs_t = [ax_ecc_vs_t]
    # Create dictionaries to store different quantities for each methods
    # as we loop over all the eccentricity and methods.
    model_eccs = {}  # To keep track of eob input eccentricity.
    measured_eccs_at_start = {}  # To keep track of measured ecc at the first
    # extrema.
    tmaxList = {}  # To keep track of minimum time in tref_out across all eccs.
    tminList = {}  # To keep track of maximum time in tref_out across all eccs.
    failed_eccs = {}  # To keep track of failed cases.
    failed_indices = {}  # To Keep track of failed indices
    # Initiate the dictionary with an empty list for each method
    for method in methods:
        model_eccs.update({method: []})
        measured_eccs_at_start.update({method: []})
        tmaxList.update({method: []})
        tminList.update({method: []})
        failed_eccs.update({method: []})
        failed_indices.update({method: []})
    q, chi1z, chi2z = available_param_sets[key]
    if key in list(available_param_sets.keys())[4:]:
        # For Fixed length cases, ecc works upto 0.89, therefore we change the
        # last two eccentricities to 0.85 and 0.89.
        EOBeccs[-2:] = [0.85, 0.89]
        eob_ecc_label = labelsDict['t_start']
    else:
        eob_ecc_label = labelsDict['omega_start']
    colors = cmap(np.linspace(0, 1, len(EOBeccs)))
    for idx0, ecc in tqdm(enumerate(EOBeccs), disable=args.verbose):

        # in debugging mode, skip all but debug_index:
        if args.debug_index is not None:
            if idx0 != args.debug_index:
                continue
        fileName, fileName_zero_ecc = get_file_name(key, ecc)
        kwargs = {"filepath": fileName}
        if args.verbose:
            print(f"idx={idx0}, {fileName}")
        # check if any residual method is in methods. If yes then load
        # zeroecc waveforms also.
        if "ResidualAmplitude" in methods or "ResidualFrequency" in methods:
            kwargs.update({"filepath_zero_ecc": fileName_zero_ecc,
                           "include_zero_ecc": True})
        dataDict = load_waveform(origin="EOB", **kwargs)
        tref_in = dataDict["t"]
        for idx, method in enumerate(methods):
            try:
                gwecc_dict = measure_eccentricity(
                    tref_in=tref_in,
                    dataDict=dataDict,
                    method=method,
                    extra_kwargs=extra_kwargs)
                tref_out = gwecc_dict["tref_out"]
                measured_ecc = gwecc_dict["eccentricity"]
                model_eccs[method] += [ecc]
                # Get the measured eccentricity at the first available index.
                # This corresponds to the first extrema that occurs after the
                # initial time
                measured_eccs_at_start[method] += [measured_ecc[0]]
                tmaxList[method] += [tref_out[-1]]
                tminList[method] += [tref_out[0]]
                # Add measured ecc vs time plot for each method to
                # corresponding axis.  Add only for idx0 that are multiples of
                # args.slice. This is to reduce the number of lines in the plot
                # to make each line plot visible. This is true only for this
                # plot.  For ecc at first extrema vs eob eccs, we plot for all
                # eccs.
                if idx0 % args.slice == 0 or idx0 == (len(EOBeccs) - 1):
                    ax_ecc_vs_t[idx].plot(tref_out, measured_ecc,
                                          c=colors[idx0], label=f"{ecc:.7f}")
                if idx == len(methods) - 1:
                    ax_ecc_vs_t[idx].set_xlabel(labelsDict["t_dimless"])
            except Exception as e:
                # collected failures
                failed_eccs[method] += [ecc]
                failed_indices[method] += [idx0]
                if args.debug_index is None:
                    # If verbose, print the exception message
                    if args.verbose:
                        print(f"Exception raised with the message {e}")
                else:
                    # if debugging, print entire traceback
                    raise
    # Iterate over methods to plots measured ecc at the first extrema vs eob
    # eccs collected above looping over all EOB eccs for each methods
    for idx, method in enumerate(methods):
        # Plot measured eccs at the first extrema vs eob eccs for different
        # methods
        ax_ecc_at_start.loglog(
            model_eccs[method], measured_eccs_at_start[method], label=method,
            c=colorsDict.get(method, "C0"),
            ls=lstyles.get(method, "-"),
            lw=lwidths.get(method, 1),
            alpha=lalphas.get(method, 1),
            zorder=100 if method == "Amplitude" else idx,  # make Amplitude appear on top
            marker=None if args.paper else "."  # No marker for paper
        )
        # Customize the measured ecc at the first extrema vs time plots
        ax_ecc_vs_t[idx].grid()
        if len(tmaxList[method]) >= 1:
            tmin = max(tminList[method])  # Choose the shortest
            # tmax = max(tmaxList[method])
            ymax = max(measured_eccs_at_start[method])
            ymin = min(EOBeccs)
            if key in list(available_param_sets.keys())[4:]:
                tmin = -20000
            ax_ecc_vs_t[idx].set_xlim(tmin, 0)  # tmax)
            ax_ecc_vs_t[idx].set_ylim(ymin, ymax)
        ax_ecc_vs_t[idx].set_ylabel(labelsDict["eccentricity"])
        ax_ecc_vs_t[idx].set_yscale("log")
        # Add yticks.
        ax_ecc_vs_t[idx].set_yticks(10.0**np.arange(-6.0, 1.0, 2.0))
        # Add text indicating the method used
        ax_ecc_vs_t[idx].text(0.95, 0.95, method, ha="right", va="top",
                              transform=ax_ecc_vs_t[idx].transAxes)
        # Add colorbar
        norm = mpl.colors.LogNorm(vmin=EOBeccs.min(), vmax=EOBeccs.max())
        divider = make_axes_locatable(ax_ecc_vs_t[idx])
        cax = divider.append_axes('right', size='3%', pad=0.1)
        sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
        cbar = fig_ecc_vs_t.colorbar(sm, cax=cax, orientation='vertical')
        # Set yticks on colorbar
        cbar.ax.set_yticks(10**np.arange(-7.0, 1.0))
        cbar.ax.tick_params(labelsize=8)
        cbar.set_label(
            rf"{labelsDict['eob_eccentricity']} at "
            rf"{eob_ecc_label}",
            size=10)
        if idx == 0:
            ax_ecc_vs_t[idx].set_title(
                rf"{labelsDict['q']}={q:.1f}, "
                fr"{labelsDict['chi1z']}={chi1z:.1f}, "
                rf"{labelsDict['chi2z']}={chi2z:.1f}",
                ha="center")

    # Customize measured eccs at the first extrema vs eob eccs
    ax_ecc_at_start.set_title(
        rf"{labelsDict['q']}={q:.1f}, "
        fr"{labelsDict['chi1z']}={chi1z:.1f}, "
        rf"{labelsDict['chi2z']}={chi2z:.1f}",
        ha="center")
    ax_ecc_at_start.legend(frameon=True)
    # Set major ticks
    locmaj = mpl.ticker.LogLocator(base=10, numticks=20)
    ax_ecc_at_start.xaxis.set_major_locator(locmaj)
    ax_ecc_at_start.yaxis.set_major_locator(locmaj)
    # Set minor ticks
    locmin = mpl.ticker.LogLocator(base=10.0, subs=np.arange(0.1, 1.0, 0.1),
                                   numticks=20)
    ax_ecc_at_start.xaxis.set_minor_locator(locmin)
    ax_ecc_at_start.xaxis.set_minor_formatter(mpl.ticker.NullFormatter())
    ax_ecc_at_start.yaxis.set_minor_locator(locmin)
    ax_ecc_at_start.yaxis.set_minor_formatter(mpl.ticker.NullFormatter())
    # Set grid
    ax_ecc_at_start.grid(which="major")
    ax_ecc_at_start.set_xlabel(rf"{labelsDict['eob_eccentricity']} at "
                               rf"{eob_ecc_label}")
    ax_ecc_at_start.set_ylabel(f"{labelsDict['eccentricity']} at {labelsDict['t_start_hat']}")
    ax_ecc_at_start.set_ylim(top=1.0)
    ax_ecc_at_start.set_xlim(EOBeccs[0], 1.0)

    # Save figures
    fig_ecc_at_start.savefig(
        f"{fig_ecc_at_start_name}",
        bbox_inches="tight")
    fig_ecc_vs_t.savefig(
        f"{fig_ecc_vs_t_name}",
        bbox_inches="tight")
    return failed_eccs, failed_indices


def report_failures(failed_eccs, failed_indices, methods):
    """Report failed cases."""
    if "all" in methods:
        methods = get_available_methods()
    for method in methods:
        num_failures = len(failed_eccs[method])
        print(f"================{method}============================")
        if num_failures == 0:
            print("All cases passed!")
        else:
            print(f"{num_failures} {'case' if num_failures == 1 else 'cases'} "
                  "failed.")
            df = pd.DataFrame({
                "Case indices": failed_indices[method],
                "Eccentricity": failed_eccs[method]
            })
            print(df)


if "all" in args.param_set_key:
    args.param_set_key = list(available_param_sets.keys())

#  Use fancy colors and other settings
style = "APS" if args.paper else "Notebook"
use_fancy_plotsettings(style=style)

for key in args.param_set_key:
    failed_eccs, failed_indices = plot_waveform_ecc_vs_model_ecc(
        args.method, key)
    print(f"================Failure reports for set {key}===================")
    report_failures(failed_eccs, failed_indices, args.method)



"""Regression test."""
import numpy as np
import json
import gw_eccentricity
from gw_eccentricity import load_data
from gw_eccentricity import measure_eccentricity
import glob

# locally it passs without problem but on github action, we need to set this tolerance
atol = 1e-9

def test_regression():
    """Regression test using all methods."""
    # List of all available methods
    available_methods = gw_eccentricity.get_available_methods()
    for method in available_methods:
        # Load regression data
        regression_data_files = glob.glob(f"test/regression_data/{method}_*_regression_data.json")
        for regression_data_file in regression_data_files:
            # Load the regression data
            print(regression_data_file)
            fl = open(regression_data_file, "r")
            regression_data = json.load(fl)
            fl.close()
            # waveform kwargs
            lal_kwargs = regression_data["waveform_kwargs"]
            # load waveform data
            dataDict = load_data.load_waveform(**lal_kwargs)
            extra_kwargs = regression_data["extra_kwargs"]

            # Try evaluating at times where regression data are saved
            regression_data_at_tref = regression_data["tref"]
            tref_in = regression_data_at_tref["time"]
            if not extra_kwargs["use_only_these_many_orbits"]:
                gwecc_dict = measure_eccentricity(
                    tref_in=tref_in,
                    method=method,
                    dataDict=dataDict,
                    extra_kwargs=extra_kwargs)
                ecc_ref = gwecc_dict["eccentricity"]
                meanano_ref = gwecc_dict["mean_anomaly"]
            else:
                ecc_ref = []
                meanano_ref = []
                for tref in tref_in:
                    gwecc_dict = measure_eccentricity(
                    tref_in=tref,
                    method=method,
                    dataDict=dataDict,
                    extra_kwargs=extra_kwargs)
                    ecc_ref.append(gwecc_dict["eccentricity"])
                    meanano_ref.append(gwecc_dict["mean_anomaly"])
                
            # Compare the measured data with the saved data
            print(ecc_ref, regression_data_at_tref["eccentricity"])
            np.testing.assert_allclose(
                ecc_ref, regression_data_at_tref["eccentricity"],
                atol=atol,
                err_msg="measured and saved eccentricity at saved times do not match.")
            np.testing.assert_allclose(
                meanano_ref, regression_data_at_tref["mean_anomaly"],
                atol=atol,
                err_msg="measured and saved mean anomaly at saved times do not match.")

            # Try evaluating at frequencies where regression data are saved
            regression_data_at_fref = regression_data["fref"]
            fref_in = regression_data_at_fref["frequency"]
            if not extra_kwargs["use_only_these_many_orbits"]:
                gwecc_dict = measure_eccentricity(
                    fref_in=fref_in,
                    method=method,
                    dataDict=dataDict,
                    extra_kwargs=extra_kwargs)
                ecc_ref = gwecc_dict["eccentricity"]
                meanano_ref = gwecc_dict["mean_anomaly"]
            else:
                ecc_ref = []
                meanano_ref = []
                for fref in fref_in:
                    gwecc_dict = measure_eccentricity(
                        fref_in=fref,
                        method=method,
                        dataDict=dataDict,
                        extra_kwargs=extra_kwargs)
                    ecc_ref.append(gwecc_dict["eccentricity"])
                    meanano_ref.append(gwecc_dict["mean_anomaly"])
            # Compare the measured data with the saved data
            np.testing.assert_allclose(
                ecc_ref, regression_data_at_fref["eccentricity"],
                atol=atol,
                err_msg="measured and saved eccentricity at saved frequencies do not match.")
            np.testing.assert_allclose(
                meanano_ref, regression_data_at_fref["mean_anomaly"],
                atol=atol,
                err_msg="measured and saved mean anomaly at saved frequencies do not match.")



import gw_eccentricity
from gw_eccentricity import load_data
from gw_eccentricity import measure_eccentricity
import numpy as np


def test_return_zero_if_small_ecc_failure():
    """Test that the interface works with return_zero_if_small_ecc_failure for
    waveforms with small or zero ecc.

    In certain situations, the waveform may have zero eccentricity or a very
    small eccentricity, making it difficult for the given method to identify
    any extrema. In cases where such a situation occurs, and if the user has
    set 'return_zero_if_small_ecc_failure' to `True` in the 'extra_kwargs'
    parameter, both the eccentricity and mean anomaly will be set to zero.
    """
    # The Amplitude and Frequency methods usually fail to detect any extrema
    # for eccentricities less than about 1e-3. Therefore, to test whether we
    # are setting eccentricity to zero when using these two methods, we use an
    # EccentricTD waveform with an initial eccentricity of 1e-4. However, since
    # the Residual and the Fits methods can detect extrema for the same
    # eccentricity, we instead use a quasicircular waveform to test the
    # handling of failures due to insufficient extrema for these methods.
    lal_kwargs_ecc = {"approximant": "EccentricTD",
                      "q": 3.0,
                      "chi1": [0.0, 0.0, 0.0],
                      "chi2": [0.0, 0.0, 0.0],
                      "Momega0": 0.01,
                      "ecc": 1e-4,
                      "mean_ano": 0.0,
                      "include_zero_ecc": True}
    lal_kwargs_qc = lal_kwargs_ecc.copy()
    lal_kwargs_qc.update({"approximant": "IMRPhenomT",
                          "ecc": 0.0})
    # Create the dataDict for Amplitude and Frequency method
    dataDict_ecc = load_data.load_waveform(**lal_kwargs_ecc)
    # Create the dataDict for Residual and Fits method
    dataDict_qc = load_data.load_waveform(**lal_kwargs_qc)
    available_methods = gw_eccentricity.get_available_methods()
    # The following will set ecc and mean ano to zero
    # if no extrema are found.
    extra_kwargs = {"return_zero_if_small_ecc_failure": True}

    # We want to test it with both a single reference point
    # as well as an array of reference points
    tref_in = {"scalar": -8000.0,
               "array": np.arange(-8000.0, 0.)}
    fref_in = {"scalar": 0.05,
               "array": np.arange(0.02, 0.05, 0.005)}
    for method in available_methods:
        if method in ["Amplitude", "Frequency"]:
            dataDict = dataDict_ecc
        else:
            dataDict = dataDict_qc
        # test for reference time
        for ref in tref_in:
            gwecc_dict = measure_eccentricity(
                tref_in=tref_in[ref],
                method=method,
                dataDict=dataDict,
                extra_kwargs=extra_kwargs)
            tref_out = gwecc_dict["tref_out"]
            ecc_ref = gwecc_dict["eccentricity"]
            meanano_ref = gwecc_dict["mean_anomaly"]
            np.testing.assert_allclose(
                ecc_ref, 0.0 if ref == "scalar"
                else np.zeros(len(tref_in[ref])))
            np.testing.assert_allclose(
                meanano_ref, 0.0 if ref == "scalar"
                else np.zeros(len(tref_in[ref])))
        # test for reference frequency
        for ref in fref_in:
            gwecc_dict = measure_eccentricity(
                fref_in=fref_in[ref],
                method=method,
                dataDict=dataDict,
                extra_kwargs=extra_kwargs)
            fref_out = gwecc_dict["fref_out"]
            ecc_ref = gwecc_dict["eccentricity"]
            meanano_ref = gwecc_dict["mean_anomaly"]
            np.testing.assert_allclose(
                ecc_ref, 0.0 if ref == "scalar"
                else np.zeros(len(fref_in[ref])))
            np.testing.assert_allclose(
                meanano_ref, 0.0 if ref == "scalar"
                else np.zeros(len(fref_in[ref])))



"""Python wrapper for SEOBNRv4EHM model by Toni."""
import lal
import lalsimulation
import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline
import os
import h5py

LALMTSUNSI = 4.925491025543575903411922162094833998e-6
# (* Subscript[M, \[CircleDot]][s] = G/c^3Subscript[M, \[CircleDot]][kg]
# ~ 4.93 10^-6s is the geometrized solar mass in seconds.
# [LAL_MTSUN_SI] defined in  lal/src/std/LALConstants.h *)
LALCSI = 299792458.0
# (* LAL_C_SI 299792458e0 /**< Speed of light in vacuo, m s^-1 *)
LALPCSI = 3.085677581491367e16
# (* LAL_PC_SI 3.085677581491367278913937957796471611e16 /**< Parsec, m * *)


def dlm(ell, m, theta):
    """Wigner d function.

    parameters
    ----------
    ell: lvalue
    m: mvalue
    theta: theta angle, e. g in GW, inclination angle iota

    Returns:
    value of d^{ell m}(theta)
    """
    kmin = max(0, m-2)
    kmax = min(ell+m, ell-2)
    d = 0
    for k in range(kmin, kmax+1):
        numerator = np.sqrt(
            float(np.math.factorial(ell+m)
                  * np.math.factorial(ell-m)
                  * np.math.factorial(ell+2)
                  * np.math.factorial(ell-2)))
        denominator = (
            np.math.factorial(k-m+2)
            * np.math.factorial(ell+m-k)
            * np.math.factorial(ell-k-2))
        d += (((-1)**k/np.math.factorial(k))
              * (numerator/denominator)
              * (np.cos(theta/2))**(2*ell+m-2*k-2)
              * (np.sin(theta/2))**(2*k-m+2))
    return d


def ylm(ell, m, theta, phi):
    """Spin weighted spherical harmonics.

    parameters:
    -----------
    ell: lvalue
    m: mvalue
    theta: theta angle, e. g in GW, inclination angle iota
    phi: phi angle, e. g. in GW, orbital phase

    Returns:
    --------
    ylm_s(theta, phi)
    """
    return (np.sqrt((2*ell + 1)/(4*np.pi))
            * dlm(ell, m, theta)
            * np.exp(1j * m * phi))


def AmpPhysicaltoNRTD(ampphysical, M, dMpc):
    return ampphysical*dMpc*1000000*LALPCSI/(LALCSI*(M*LALMTSUNSI))


def compute_freqInterp(time, hlm):
    philm = np.unwrap(np.angle(hlm))

    intrp = InterpolatedUnivariateSpline(time, philm)
    omegalm = intrp.derivative()(time)

    return omegalm


def SectotimeM(seconds, M):
    return seconds/(M*LALMTSUNSI)


def get_sphtseries(q: float, chi1: float, chi2: float,
                   eccentricity: float,
                   eccentric_anomaly: float,
                   f_min: float, M_fed: float, delta_t: float,
                   EccIC: int,
                   dMpc: float,
                   approx: str):

    # Some internal settings of the model
    HypPphi0, HypR0, HypE0 = [0., 0, 0]
    EccFphiPNorder = 99
    EccFrPNorder = 99
    EccWaveformPNorder = 16
    EccBeta = 0.09
    Ecct0 = 100

    EccPNFactorizedForm = EccNQCWaveform = EccPNRRForm = EccPNWfForm = 1
    EccAvNQCWaveform = 1
    EcctAppend = 40
    # EccIC = 0
    # eccentric_anomaly = 0.0

    m1 = q/(1+q)*M_fed
    m2 = 1/(1+q)*M_fed
    dist = dMpc*(1e6*lal.PC_SI)
    SpinAlignedVersion = 41
    nqcCoeffsInput = lal.CreateREAL8Vector(50)

    sphtseries, dyn, dynHi = lalsimulation.SimIMRSpinAlignedEOBModesEcc_opt(
        delta_t,
        m1*lal.MSUN_SI,
        m2*lal.MSUN_SI,
        f_min,
        dist,
        chi1,
        chi2,
        eccentricity,
        eccentric_anomaly,
        SpinAlignedVersion, 0., 0., 0., 0., 0., 0., 0., 0., 1., 1.,
        nqcCoeffsInput, 0,
        EccFphiPNorder, EccFrPNorder, EccWaveformPNorder,
        EccPNFactorizedForm, EccBeta, Ecct0, EccNQCWaveform,
        EccPNRRForm, EccPNWfForm, EccAvNQCWaveform,
        EcctAppend, EccIC, HypPphi0, HypR0, HypE0)
    return sphtseries, dyn, dynHi


def get_modes(
        q=1,
        chi1=0,
        chi2=0,
        eccentricity=1e-5,
        eccentric_anomaly=0,
        f_min=20,
        M_fed=50,
        delta_t=1/2048,
        EccIC=0,
        dMpc=500,
        approx="SEOBNRv4EHM_opt",
        physical_units=True,
        save=True):
    """Get the hlm modes."""
    mode_dir = "./seobmode_data"
    if not os.path.exists(mode_dir):
        os.mkdir(mode_dir)
    mode_file = (f"{mode_dir}/q_{q}_ecc_{eccentricity}_ano_"
                 f"{eccentric_anomaly}_f_min_{f_min}_chi1_{chi1}_chi2_{chi2}_"
                 f"M_fed_{M_fed}_srate_{int(1/delta_t)}_EccIC_{EccIC}_"
                 f"dMpc_{dMpc}_approx_{approx}_phys_units_{physical_units}.h5")
    hlm = {}
    mode_list = [[2, 2],
                 [3, 3],
                 [4, 4],
                 [5, 5]]
    if os.path.exists(mode_file):
        mode_data = h5py.File(mode_file, "r")
        time_array = mode_data["time"][:]
        for ell, m in mode_list:
            hlm[ell, m] = mode_data[f"strain/({ell, m})"][:]
        mode_data.close()
    else:
        sphtseries, dyn, dynHi = get_sphtseries(
            q, chi1, chi2,
            eccentricity,
            eccentric_anomaly,
            f_min,
            M_fed,
            delta_t,
            EccIC,
            dMpc,
            approx)

        # 55 mode
        modeL = sphtseries.l
        modeM = sphtseries.m
        h55 = sphtseries.mode.data.data  # This is h_55
        hlm[modeL, modeM] = h55  # AmpPhysicaltoNRTD(h55, M_fed, dMpc)
        # hlm[modeL, -modeM] = ((-1)**modeL) * np.conjugate(h55)

        # 44 mode
        modeL = sphtseries.next.l
        modeM = sphtseries.next.m
        h44 = sphtseries.next.mode.data.data  # This is h_44
        hlm[modeL, modeM] = h44  # AmpPhysicaltoNRTD(h44, M_fed, dMpc)
        # hlm[modeL, -modeM] = ((-1)**modeL) * np.conjugate(h44)

        # 21 mode
        # modeL = sphtseries.next.next.l
        # modeM = sphtseries.next.next.m
        # h21 = sphtseries.next.next.mode.data.data  # This is h_21
        # hlm[modeL, modeM] = AmpPhysicaltoNRTD(h21, M_fed, dMpc)
        # hlm[modeL, -modeM] = ((-1)**modeL) * np.conjugate(h21)

        # 33 mode
        modeL = sphtseries.next.next.next.l
        modeM = sphtseries.next.next.next.m
        h33 = sphtseries.next.next.next.mode.data.data  # This is h_33
        hlm[modeL, modeM] = h33  # AmpPhysicaltoNRTD(h33, M_fed, dMpc)
        # hlm[modeL, -modeM] = ((-1)**modeL) * np.conjugate(h33)

        # 22 mode
        modeL = sphtseries.next.next.next.next.l
        modeM = sphtseries.next.next.next.next.m
        h22 = sphtseries.next.next.next.next.mode.data.data  # This is h_22
        hlm[modeL, modeM] = h22  # AmpPhysicaltoNRTD(h22, M_fed, dMpc)
        # hlm[modeL, -modeM] = ((-1)**modeL) * np.conjugate(h22)

        time_array = np.arange(0, len(h22)*delta_t, delta_t)
        if not physical_units:
            time_array = SectotimeM(time_array, M_fed)
            for key in hlm.keys():
                hlm[key] = AmpPhysicaltoNRTD(hlm[key], M_fed, dMpc)

        if save:
            mode_data = h5py.File(mode_file, "w")
            mode_data["time"] = time_array
            for ell, m in mode_list:
                mode_data[f"strain/({ell, m})"] = hlm[ell, m]
            mode_data.close()

    return time_array, hlm



"""
Find pericenters and apocenters using Residual Amplitude.

Part of Eccentricity Definition project.
Md Arif Shaikh, Mar 29, 2022
"""
from .eccDefinitionUsingAmplitude import eccDefinitionUsingAmplitude


class eccDefinitionUsingResidualAmplitude(eccDefinitionUsingAmplitude):
    """Measure eccentricity by finding extrema from residual amplitude."""

    def __init__(self, *args, **kwargs):
        """Init for eccDefinitionUsingResidualAmplitude class.

        parameters:
        ----------
        dataDict: Dictionary containing the eccentric and quasi
        circular waveform data.
        For residual amplitude method we need quasi-circular modes
        in additionn to the eccentric modes. Provide it as a dictionary
        for the key `hlm_zeroecc` and `t_zeroecc` in the dataDict dictionary.
        Keys for the modes in the mode dictionary should be
        of the form `(l, m)`.
        """
        super().__init__(*args, **kwargs)
        self.method = "ResidualAmplitude"
        self.label_for_data_for_finding_extrema = self.get_label_for_plots("res_amp")

    def check_and_raise_zeroecc_data_not_found(self, method):
        """Raise exception if zeroecc data not found for Residual method.

        For Residual methods, waveform data of the quasicircular counterpart
        is required to compute residual amplitude/frequency which is then
        used to find the locations of the pericenters/apocenters.

        In the dataDict, these are provided using the keys
        - t_zeroecc: 1d array of times of quasicircular waveform.
        - hlm_zeroecc: Dictionary of modes of the quasicircular waveform.
        For more details on the format of the dataDict, see documentation
        of gw_eccentricity.measure_eccentricity.
        """
        # suggest what data to provide when zeroecc amplitude/omega is missing
        # for Residual methods
        suggested_keys = {
            "t_zeroecc": ["t_zeroecc"],
            "amplm_zeroecc": ["amplm_zeroecc", "hlm_zeroecc"],
            "omegalm_zeroecc": ["omegalm_zeroecc", "hlm_zeroecc", "phaselm_zeroecc"]}
        for k in suggested_keys:
            if k not in self.dataDict:
                raise Exception(f"Method {method} must have zeroecc data in "
                                f"dataDict. {k} data not found. "
                                "At least one of the following data should be "
                                f"provided: {suggested_keys[k]}")

    def get_data_for_finding_extrema(self):
        """Get the data for extrema finding."""
        self.check_and_raise_zeroecc_data_not_found("ResidualAmplitude")
        return self.res_amp_gw
